# Copyright (c) Alibaba, Inc. and its affiliates.

import json
from typing import Dict

import datasets
from datasets import Features, Value

from adaseq.data.constant import PAD_LABEL

from .base import CustomDatasetBuilder


class NamedEntityRecognitionDatasetBuilderConfig(datasets.BuilderConfig):
    """BuilderConfig for Named Entity Recognition datasets"""

    def __init__(self, data_dir=None, data_files=None, **corpus_config):
        super().__init__(data_dir=data_dir, data_files=data_files)
        self.corpus_config = corpus_config


class NamedEntityRecognitionDatasetBuilder(CustomDatasetBuilder):
    """Builder for entity typing datasets.

    features:
        id: string, data record id.
        tokens: list[str] input tokens.
        spans: List[Dict],  mentions like: [{'start': 0, 'end': 2, 'type': 'PER'}]
        mask: bool, mention mask.
    """

    BUILDER_CONFIG_CLASS = NamedEntityRecognitionDatasetBuilderConfig

    def stub():  # noqa: D102
        pass

    def _info(self):
        info = datasets.DatasetInfo(
            features=Features(
                {
                    'id': Value('string'),
                    'tokens': [Value('string')],
                    'spans': [
                        {
                            'start': Value('int32'),  # close
                            'end': Value('int32'),  # open
                            'type': Value('string'),
                        }
                    ],
                    'mask': [Value('bool')],
                }
            )
        )
        return info

    def _generate_examples(self, filepath):
        """
        NER reader supports:
        1. data_type: conll
            ```
            duck B-PER
            duck I-PER
            duck O
            ```

        2. data_type: json_tags
        ```
        {
            'text': 'duck duck duck duck',
            'labels': ['B-PER', 'O', ...]
        }
        ```

        3. data_type: json_spans
        ```
        {
            'text': 'duck duck',
            'spans': [
                {'start': 0, 'end': 1, 'type': 'PER'},
                ...
            ]
        ```

        4. data_type: cluener
        ```
        {
            'text': 'duck duck',
            'label': {
                'LOC': [[0, 1], ...]
            }
        }
        ```
        """
        corpus_config = self.config.corpus_config
        if corpus_config['data_type'] == 'conll':
            return self._load_conll_file(filepath, corpus_config)
        elif corpus_config['data_type'] == 'json_tags':
            return self._load_json_tags_file(filepath, corpus_config)
        elif corpus_config['data_type'] == 'json_spans':
            return self._load_json_spans_file(filepath, corpus_config)
        elif corpus_config['data_type'] == 'cluener':
            return self._load_cluener_file(filepath, corpus_config)
        else:
            raise ValueError('Unknown corpus format type [%s]' % corpus_config['data_type'])

    @classmethod
    def _load_conll_file(cls, file_path, corpus_config: Dict):
        delimiter = corpus_config.get('delimiter', None)
        with open(file_path, encoding='utf-8') as f:
            guid = 0
            tokens = []
            labels = []
            for line in f:
                if line.startswith('-DOCSTART-') or line == '' or line == '\n':
                    if tokens:
                        spans = cls._labels_to_spans(labels)
                        mask = cls._labels_to_mask(labels)
                        yield guid, {
                            'id': str(guid),
                            'tokens': tokens,
                            'spans': spans,
                            'mask': mask,
                        }
                        guid += 1
                        tokens = []
                        labels = []
                else:
                    splits = line.split(delimiter)
                    tokens.append(splits[0])
                    labels.append(splits[-1].rstrip())
            if tokens:
                spans = cls._labels_to_spans(labels)
                mask = cls._labels_to_mask(labels)
                yield guid, {'id': str(guid), 'tokens': tokens, 'spans': spans, 'mask': mask}

    @classmethod
    def _load_json_tags_file(cls, filepath, corpus_config):
        tags_key = corpus_config.get('tags_key', 'labels')
        text_key = corpus_config.get('text_key', 'text')
        tokenizer = corpus_config.get('tokenizer', 'char')
        with open(filepath, encoding='utf-8') as f:
            guid = 0
            for line in f:
                example = json.loads(line)
                text = example[text_key]
                if isinstance(text, list):
                    tokens = text
                elif isinstance(text, str):
                    if tokenizer == 'char':
                        tokens = list(text)
                    elif tokenizer == 'blank':
                        tokens = text.split(' ')
                    else:
                        raise NotImplementedError
                else:
                    raise ValueError('Unsupported text input.')
                labels = example[tags_key]
                assert len(tokens) == len(labels)
                spans = cls._labels_to_spans(labels)
                mask = cls._labels_to_mask(labels)
                yield guid, {'id': str(guid), 'tokens': tokens, 'spans': spans, 'mask': mask}
                guid += 1

    @classmethod
    def _load_json_spans_file(cls, filepath, corpus_config):
        # {'text': 'aaa', 'labels': [{'start': 0, 'end': 1, type: 'LOC'}, ...]}
        # {'tokens': ['a', 'aa', ...], 'spans': [{'start': 0, 'end': 1, type: 'LOC'}, ...]}
        spans_key = corpus_config.get('spans_key', 'spans')
        text_key = corpus_config.get('text_key', 'text')
        tokenizer = corpus_config.get('tokenizer', 'char')
        is_end_included = corpus_config.get('is_end_included', False)

        with open(filepath, encoding='utf-8') as f:
            guid = 0
            for line in f:
                if line.strip() == '':
                    continue
                example = json.loads(line)
                text = example[text_key]
                if isinstance(text, list):
                    tokens = text
                elif isinstance(text, str):
                    if tokenizer == 'char':
                        tokens = list(text)
                    elif tokenizer == 'blank':
                        tokens = text.split(' ')
                    else:
                        raise RuntimeError
                else:
                    raise ValueError('Unsupported text input.')
                spans = []
                for span in example[spans_key]:
                    if is_end_included:
                        span['end'] += 1
                    if 'word' in span:
                        del span['word']
                    spans.append(span)
                mask = [True] * len(tokens)
                yield guid, {'id': str(guid), 'tokens': tokens, 'spans': spans, 'mask': mask}
                guid += 1

    @classmethod
    def _load_cluener_file(cls, filepath, corpus_config):
        is_end_included = corpus_config.get('is_end_included', False)

        with open(filepath, encoding='utf-8') as f:
            guid = 0
            for line in f:
                example = json.loads(line)
                text = example['text']
                if isinstance(text, list):
                    tokens = text
                elif isinstance(text, str):
                    if corpus_config['tokenizer'] == 'char':
                        tokens = list(text)
                    elif corpus_config['tokenizer'] == 'blank':
                        tokens = text.split(' ')
                    else:
                        raise NotImplementedError
                else:
                    raise NotImplementedError
                entities = list()
                for entity_type, span_list in example['label'].items():
                    for name, span in span_list.items():
                        end_offset = int(is_end_included)
                        span = dict(start=span[0][0], end=span[0][1] + end_offset, type=entity_type)
                        entities.append(span)
                mask = [True] * len(tokens)
                yield guid, {'id': str(guid), 'tokens': tokens, 'spans': entities, 'mask': mask}
                guid += 1

    @classmethod
    def _labels_to_spans(cls, labels):
        spans = []
        in_entity = False
        start = -1
        for i, tag in enumerate(labels):
            # fix label error
            if tag[0] in 'IE' and not in_entity:
                tag = 'B' + tag[1:]
            if tag[0] in 'BS':
                if i + 1 < len(labels) and labels[i + 1][0] in 'IE':
                    start = i
                else:
                    spans.append(dict(start=i, end=i + 1, type=tag[2:]))
            elif tag[0] in 'IE':
                if i + 1 >= len(labels) or labels[i + 1][0] not in 'IE':
                    assert start >= 0, 'Invalid label sequence found: {}'.format(labels)
                    spans.append(dict(start=start, end=i + 1, type=tag[2:]))
                    start = -1
            if tag[0] in 'B':
                in_entity = True
            elif tag[0] in 'OES':
                in_entity = False
        return spans

    @classmethod
    def _labels_to_mask(cls, labels):
        mask = []
        for label in labels:
            mask.append(label != PAD_LABEL)
        return mask



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict

from adaseq.metainfo import DataCollators

from .base import DATA_COLLATORS, DataCollatorWithPadding


@DATA_COLLATORS.register_module(module_name=DataCollators.twostage_data_collator)
class TwostageDataCollatorWithPadding(DataCollatorWithPadding):
    """Padding method for two stage ner dataset."""

    def padding(self, batch: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        """
        Padding a batch. In addition to the fields padded by base class DataCollatorWithPadding,
        'mention_boundary', 'type_ids', 'mention_mask', 'ident_ids', 'ident_mask' are padded here.
        """

        max_length = max(len(i) for i in batch['ident_ids'])
        for i in range(len(batch['ident_ids'])):
            difference = max_length - len(batch['ident_ids'][i])
            if difference > 0:
                batch['ident_ids'][i] = batch['ident_ids'][i] + [-100] * difference

        max_span_count = max([len(x[0]) for x in batch['mention_boundary']])
        for i in range(len(batch['mention_boundary'])):
            difference = max_span_count - len(batch['mention_boundary'][i][0])
            if difference > 0:
                batch['mention_boundary'][i][0] = batch['mention_boundary'][i][0] + [0] * difference
                batch['mention_boundary'][i][1] = batch['mention_boundary'][i][1] + [0] * difference
                batch['type_ids'][i] = batch['type_ids'][i] + [-100] * difference
                batch['mention_mask'][i] = batch['mention_mask'][i] + [0] * difference
        return batch



# Copyright (c) Alibaba, Inc. and its affiliates.
DEFAULT_CONFIG = {
    'train': {
        'dataloader': {'workers_per_gpu': 0},
        'hooks': [
            {'type': 'EvaluationHook'},
            {'type': 'CheckpointHook'},
            {
                'type': 'BestCkptSaverHook',
                'save_file_name': 'best_model.pth',
                'metric_key': 'f1',
                'save_optimizer': False,
                'restore_best': True,
            },
            {'type': 'TextLoggerHook', 'interval': 50},
            {'type': 'IterTimerHook'},
        ],
    },
    'evaluation': {'dataloader': {'workers_per_gpu': 0, 'shuffle': False}},
}



# Copyright (c) Alibaba, Inc. and its affiliates.
import random
from typing import Any, Dict, List, Union

import numpy as np
from modelscope.preprocessors.builder import PREPROCESSORS
from modelscope.utils.constant import Fields

from adaseq.metainfo import Preprocessors

from .nlp_preprocessor import NLPPreprocessor


@PREPROCESSORS.register_module(Fields.nlp, module_name=Preprocessors.pretraining_preprocessor)
class PretrainingPreprocessor(NLPPreprocessor):
    """Preprocessor for pretraining.
    span targets are processed into mention_boundary, type_ids, ident_ids.
    examples:
        span: {'start':1, 'end':2, 'type': ['PER']}
        processed: {'mention_boundary': [[1], [2]], 'type_ids':[1], 'ident_ids': ['S-SPAN']]}
    """

    def __init__(self, model_dir: str, **kwargs) -> None:

        super().__init__(model_dir, return_offsets=True, **kwargs)

        label_to_id = kwargs.pop('label_to_id', None)
        labels = kwargs.pop('labels', None)
        if label_to_id is not None:
            self.typing_label_to_id = label_to_id
        elif labels is not None:
            self.typing_label_to_id = self.make_label_to_id(labels)
        else:
            raise ValueError('Must have one of `labels` or `label_to_id`')
        self.typing_id_to_label = {
            v: k for k, v in sorted(self.typing_label_to_id.items(), key=lambda x: x[1])
        }
        # make sure they are aligned.
        assert len(self.typing_id_to_label) not in self.typing_id_to_label
        assert len(self.typing_id_to_label) - 1 in self.typing_id_to_label

        self.ident_label_to_id = {'O': 0, 'B-SPAN': 1, 'I-SPAN': 2, 'E-SPAN': 3, 'S-SPAN': 4}
        self.ident_id_to_label = {0: 'O', 1: 'B-SPAN', 2: 'I-SPAN', 3: 'E-SPAN', 4: 'S-SPAN'}

    def __call__(self, data: Union[str, List, Dict]) -> Dict[str, Any]:
        """prepare inputs for Pretraining model"""
        prompt_type = None
        if isinstance(data, str):
            prompt_sep_pos = data.find('[SEP]')
            if prompt_sep_pos > 0:
                prompt_type = data[prompt_sep_pos + len('[SEP]') :]
                data = data[:prompt_sep_pos]
            data = {'text': data}
        if isinstance(data, List):
            try:
                prompt_sep_pos = data.index('[SEP]')
            except ValueError:
                prompt_sep_pos = -1
            if prompt_sep_pos > 0:
                prompt_type = data[prompt_sep_pos + 1 :]
                data = data[:prompt_sep_pos]
            data = {'tokens': data}
        output = super().__call__(data)
        if prompt_type is not None:
            self.__generate_prompt_data(data, output, prompt_type)
        return output

    def __generate_prompt_data(
        self,
        data: Union[str, List, Dict],
        output: Dict[str, Any],
        prompt_type: Union[str, List] = None,
    ):
        testing_mode = False
        # collect types
        if prompt_type is None:
            present_types = set([])
            for span in data['spans']:
                present_types.add(span['type'][0])
            # select target type
            all_type = self.typing_label_to_id.keys()
            threshold = 1.0 / len(all_type)
            while len(all_type) > len(present_types):
                tmp_type = random.sample(all_type, 1)[0]
                if tmp_type not in present_types and random.random() <= threshold:
                    present_types.add(tmp_type)
                    break
            selected_type = random.sample(present_types, 1)[0]
            prompt_type = selected_type
        else:
            testing_mode = True
        # construct data
        prompt_token_ids = self.encode_tokens(list(prompt_type))['input_ids']
        # skip [CLS], in inference model, nlp_preprocessor will add batch dimssion
        if testing_mode:
            prompt_input_ids = list(output['tokens']['input_ids'][0]) + prompt_token_ids[1:]
        else:
            prompt_input_ids = output['tokens']['input_ids'] + prompt_token_ids[1:]
        output['prompt_input_ids'] = prompt_input_ids
        output['prompt_input_mask'] = [True] * len(prompt_input_ids)
        if testing_mode:
            output['prompt_input_ids'] = np.expand_dims(np.array(output['prompt_input_ids']), 0)
            output['prompt_input_mask'] = np.expand_dims(
                np.array(output['prompt_input_mask']), False
            )
            return

        prompt_target_label_ids = []
        prompt_target_label_ids.extend(output['ident_ids'])

        for span in data['spans']:
            if span['type'][0] != selected_type:
                for pos in range(span['start'], span['end']):
                    if pos < len(prompt_target_label_ids):
                        prompt_target_label_ids[pos] = self.ident_label_to_id[
                            'O'
                        ]  # mask mention of the other types
        output['prompt_target_label_ids'] = prompt_target_label_ids
        output['prompt_type'] = prompt_type



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from modelscope.models.builder import MODELS
from torch.nn import LayerNorm

from adaseq.metainfo import Models, Tasks
from adaseq.models.base import Model
from adaseq.modules.dropouts import WordDropout
from adaseq.modules.embedders import Embedder


@MODELS.register_module(Tasks.relation_extraction, module_name=Models.relation_extraction_model)
class RelationExtractionModel(Model):
    """Relation extraction model

    This model is used for relation extraction tasks.
    """

    def __init__(
        self,
        id_to_label: Dict[int, str],
        embedder: Union[Embedder, str],
        dropout: float = 0.0,
        word_dropout: bool = False,
        multiview: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        **kwargs
    ):
        super(RelationExtractionModel, self).__init__(**kwargs)
        self.id_to_label = id_to_label
        self.num_labels = len(id_to_label)
        if isinstance(embedder, Embedder):
            self.embedder = embedder
        else:
            self.embedder = Embedder.from_config(embedder)

        self.use_dropout = dropout > 0.0
        if self.use_dropout:
            if word_dropout:
                self.dropout = WordDropout(dropout)
            else:
                self.dropout = nn.Dropout(dropout)

        hidden_size = self.embedder.get_output_dim()
        self.linear = nn.Linear(2 * hidden_size, self.num_labels)
        self.layer_norm = LayerNorm(2 * hidden_size)

        self.loss_fn = nn.CrossEntropyLoss(reduction='mean')

        self.multiview = multiview
        self.temperature = temperature

    def _forward(self, tokens: Dict[str, Any], so_head_mask: torch.Tensor) -> torch.Tensor:
        embed = self.embedder(**tokens)

        if self.use_dropout:
            embed = self.dropout(embed)

        batch_size = so_head_mask.shape[0]
        so_emb = embed.masked_select(so_head_mask.unsqueeze(-1)).view(batch_size, -1)

        so_emb = self.layer_norm(so_emb)
        logits = self.linear(so_emb)

        return logits

    def forward(  # noqa
        self,
        tokens: Dict[str, Any],
        so_head_mask: torch.BoolTensor,
        meta: Dict[str, Any],
        label_id: Optional[torch.LongTensor] = None,
        origin_tokens: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:  # TODO: add docstring
        logits = self._forward(tokens, so_head_mask)

        if self.training:
            loss = self._calculate_loss(logits, label_id)
            if self.multiview and origin_tokens is not None:  # for multiview training
                origin_logits = self._forward(origin_tokens, so_head_mask)
                origin_loss = self._calculate_loss(origin_logits, label_id)
                cl_kl_loss = self._calculate_cl_loss(logits, origin_logits, T=self.temperature)
                loss = loss + origin_loss + cl_kl_loss
            outputs = {'logits': logits, 'loss': loss}
        else:
            predicts = self.decode(logits)
            outputs = {'logits': logits, 'predicts': predicts}

        return outputs

    def _calculate_cl_loss(self, ext_view_logits, origin_view_logits, T=1.0):
        if self.multiview:
            batch_size, num_classes = ext_view_logits.shape
            ext_view_logits = ext_view_logits.detach()
            _loss = (
                F.kl_div(
                    F.log_softmax(origin_view_logits / T, dim=-1),
                    F.softmax(ext_view_logits / T, dim=-1),
                    reduction='none',
                )
                * T
                * T
            )
            loss = _loss.sum() / batch_size
        else:
            loss = 0.0
        return loss

    def _calculate_loss(self, logits, targets):
        loss = self.loss_fn(logits, targets)
        return loss

    def decode(self, logits):  # noqa
        predicts = logits.argmax(-1)
        return predicts



# Copyright (c) Alibaba, Inc. and its affiliates.
from os import path as osp
from typing import Dict, Union

import torch.nn as nn
from modelscope.utils.config import Config, ConfigDict
from modelscope.utils.registry import Registry, build_from_cfg

from adaseq.metainfo import Embedders

EMBEDDERS = Registry('embedders')


def build_embedder(cfg: ConfigDict, default_args: dict = None):
    """Build embedder from config dict

    Args:
        cfg (:obj:`ConfigDict`): config dict for embedder object
        default_args (dict): default initialization arguments

    Returns:
        embedder (:obj:`Embedder`): an embedder instance
    """
    return build_from_cfg(cfg, EMBEDDERS, group_key='default', default_args=default_args)


class Embedder(nn.Module):
    """
    The embedder base class for encoding input_ids to hidden-states
    """

    def get_output_dim(self) -> int:
        """
        Get the output embedding dim.
        """
        raise NotImplementedError

    @classmethod
    def from_config(cls, cfg_dict_or_path: Union[str, Dict, Config] = None, **kwargs) -> 'Embedder':
        """Build embedder instance from config"""
        if isinstance(cfg_dict_or_path, str) and osp.isfile(cfg_dict_or_path):
            cfg = Config.from_file(cfg_dict_or_path).model.embedder
        elif isinstance(cfg_dict_or_path, (dict, Config)):
            cfg = cfg_dict_or_path
        else:
            cfg = {}

        if 'type' not in cfg:
            if 'model_name_or_path' in cfg:
                cfg['type'] = Embedders.transformer_embedder
            else:
                cfg['type'] = None
        if 'type' in kwargs:
            cfg['type'] = kwargs.pop('type')

        if 'model_name_or_path' not in cfg:
            cfg['model_name_or_path'] = None
        if 'model_name_or_path' in kwargs:
            cfg['model_name_or_path'] = kwargs.pop('model_name_or_path')
            if cfg['type'] is None:
                cfg['type'] = Embedders.transformer_embedder

        if cfg['type'] is not None and cfg['type'] in EMBEDDERS.modules['default']:
            return build_embedder(cfg, default_args=kwargs)
        else:
            raise ValueError('Unsupported embedder type: %s', cfg['type'])



# Copyright (c) Alibaba, Inc. and its affiliates.
NON_ENTITY_LABEL = 'O'
NON_ENTITY_LABEL_ID = 0
PAD_LABEL = 'X'
PAD_LABEL_ID = -100

OBJECT_START_TOKEN = '<O>'
SUBJECT_START_TOKEN = '<E>'
NONE_REL_LABEL = 'None'
NONE_REL_LABEL_ID = 0

PARTIAL_LABEL = 'P'
PARTIAL_LABEL_ID = -1



# Copyright (c) Alibaba, Inc. and its affiliates.
from abc import abstractmethod
from os import path as osp
from typing import Dict, Union

import torch
import torch.nn as nn
from modelscope.utils.config import Config, ConfigDict
from modelscope.utils.registry import Registry, build_from_cfg
from transformers import AutoModel

from adaseq.utils.common_utils import has_keys

DECODERS = Registry('decoders')


def build_decoder(cfg: ConfigDict, default_args: dict = None):
    """Build decoder from config dict

    Args:
        cfg (:obj:`ConfigDict`): config dict for decoder object
        default_args (dict): default initialization arguments

    Returns:
        encoder (:obj:`Decoder`): an decoder instance
    """
    return build_from_cfg(cfg, DECODERS, group_key='default', default_args=default_args)


class Decoder(nn.Module):
    """
    The decoder base class for downstream tasks
    """

    def __init__(self, **kwargs):
        super().__init__()

    @classmethod
    def instantiate(cls, model_name_or_path=None, config=None, **kwargs):
        """Instantiate a decoder class, the __init__ method will be called by default.

        Args:
            model_name_or_path: The model_name_or_path parameter used to initialize a decoder.
            config: The config dict read from the caller's cfg_dict_or_path parameter.
            **kwargs: Extra parameters.

        Returns: A Decoder instance.
        """
        return cls(model_name_or_path=model_name_or_path, config=config, **kwargs)

    @abstractmethod
    @torch.jit.export
    def decode(self, logits, mask=None, **kwargs):
        """Decode logits"""
        raise NotImplementedError

    @classmethod
    def from_config(
        cls, model_name_or_path: str = None, cfg_dict_or_path: Union[str, Dict] = None, **kwargs
    ):
        """Build an decoder subclass.

        Args:
            model_name_or_path: The model_name_or_path parameter used to initialize a decoder.
            cfg_dict_or_path: The extra config file or the extra config dict.
            **kwargs:
                decoder_type: Same with cfg.model.decoder.type

        Returns: An Decoder instance.
        """
        assert (
            model_name_or_path is not None or cfg_dict_or_path is not None
        ), 'Either the model or the cfg information should be passed in from the parameters.'

        if cfg_dict_or_path is not None:
            if isinstance(cfg_dict_or_path, str) and osp.isfile(cfg_dict_or_path):
                cfg = Config.from_file(cfg_dict_or_path)
            elif isinstance(cfg_dict_or_path, (dict, Config)):
                cfg = cfg_dict_or_path
            else:
                raise ValueError(
                    'Please pass a correct cfg dict, which should be a reachable file or a dict.'
                )
        elif model_name_or_path is not None and osp.exists(
            osp.join(model_name_or_path, 'config.json')
        ):
            cfg = Config.from_file(osp.join(model_name_or_path, 'config.json'))
        else:
            cfg = {}

        type = None
        if 'decoder_type' in kwargs:
            type = kwargs.pop('decoder_type')
        elif has_keys(cfg, 'model', 'decoder', 'type'):
            type = cfg['model']['decoder']['type']
        elif 'model' not in cfg and 'type' in cfg:
            type = cfg['type']
        elif 'model' not in cfg and 'model_type' in cfg:
            type = cfg['model_type']

        if model_name_or_path is None and has_keys(cfg, 'model', 'decoder', 'model_name_or_path'):
            model_name_or_path = cfg['model']['decoder']['model_name_or_path']

        if type is not None and type in DECODERS.modules['default']:
            return build_decoder(
                type,
                model_name_or_path=model_name_or_path,
                config=cfg if len(cfg) > 0 else None,
                **kwargs
            )
        else:
            assert model_name_or_path is not None, (
                'Model is not found in registry, '
                'so it is considered a huggingface backbone '
                'and the model_name_or_path param should not be None'
            )
            return AutoModel.from_pretrained(model_name_or_path, **kwargs)



# Copyright (c) Alibaba, Inc. and its affiliates.
import logging
import os
import os.path as osp
from abc import abstractmethod
from typing import Any, Callable, Dict, Optional, Union

import torch
import torch.nn as nn
from modelscope.hub.snapshot_download import snapshot_download
from modelscope.models.base.base_model import Model as MsModel
from modelscope.models.builder import build_model
from modelscope.utils.checkpoint import save_checkpoint
from modelscope.utils.config import Config, ConfigDict
from modelscope.utils.constant import DEFAULT_MODEL_REVISION, ModelFile

logger = logging.getLogger(__name__)


class Model(nn.Module, MsModel):  # TODO 继承 modelscope model
    """
    The model base class
    """

    def __init__(self, model_dir: str = None, **kwargs):
        super().__init__()
        MsModel.__init__(self, model_dir=model_dir, **kwargs)

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)

        def new_init(self, init=cls.__init__, *args, **kwargs):
            init(self, *args, **kwargs)
            self.post_init()

        cls.__init__ = new_init

    def post_init(self):
        """Run something after __init__ in subclass

        All derived model instances will try to load checkpoint from model_dir after __init__.
        Useful when initializing pipeline from model_id.
        """
        self.load_model_ckpt()

    def __call__(self, *args, **kwargs) -> Dict[str, Any]:  # noqa: D102
        # use `nn.Module.__call__` rather than `MsModel.__call__`
        return self.postprocess(super()._call_impl(*args, **kwargs))

    def load_model_ckpt(self, model_dir: str = None) -> None:
        """Try to load model checkpoint from model_dir/pytorch_model.bin"""
        if model_dir is None:
            model_dir = self.model_dir
        if model_dir is not None:
            model_ckpt = os.path.join(model_dir, ModelFile.TORCH_MODEL_BIN_FILE)
            self.load_state_dict(torch.load(model_ckpt, map_location=torch.device('cpu')))

    @abstractmethod
    def forward(
        self, tokens: Dict[str, Any], meta: Optional[Dict[str, Any]] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Run the forward pass for a model.

        Args:
            tokens (Dict[str, Any]): inputs to the model
            meta (Dict[str, Any]): raw inputs
            **kwargs: other arguments

        Returns:
            outputs (Dict[str, Any]): outputs from the model
        """
        raise NotImplementedError

    @classmethod
    def from_config(cls, cfg_dict_or_path: Optional[Union[str, Dict, Config]] = None, **kwargs):
        """Instantiate a model from config dict or path.

        Args:
            cfg_dict_or_path (Optional[Union[str, Dict, Config]]): config dict or path
            **kwargs: other arguments

        Returns:
            model (:obj:`Model`): a model instance
        """

        if isinstance(cfg_dict_or_path, str) and osp.isfile(cfg_dict_or_path):
            cfg = Config.from_file(cfg_dict_or_path)
        elif isinstance(cfg_dict_or_path, (dict, Config)):
            cfg = cfg_dict_or_path
        else:
            cfg = dict(model=dict(), task=None)

        task: str = cfg['task']
        if 'task' in kwargs:
            task = kwargs.pop('task')
        model_config: Dict = cfg['model']

        if 'type' in kwargs:
            model_config['type'] = kwargs.pop('type')

        if 'type' not in model_config:
            raise ValueError(
                'Please pass a correct cfg dict, which should be a reachable file or a dict.'
            )

        model = build_model(model_config, task_name=task, default_args=kwargs)
        cfg['framework'] = 'pytorch'
        cfg['model'].update(kwargs)  # type: ignore
        setattr(model, 'cfg', cfg)  # follow `MsModel.save_pretrained`
        return model

    def save_pretrained(
        self,
        target_folder: Union[str, os.PathLike],
        save_checkpoint_names: str = ModelFile.TORCH_MODEL_BIN_FILE,
        save_function: Callable = save_checkpoint,
        config: Optional[dict] = None,
        manual_call: bool = False,
        **kwargs,
    ) -> None:
        """`MsModel.save_pretrained`"""
        if not manual_call:
            return  # prevent modelscope call this method in training stage.
        super().save_pretrained(
            target_folder, save_checkpoint_names, save_function, config, **kwargs
        )

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        revision: Optional[str] = DEFAULT_MODEL_REVISION,
        cfg_dict: Config = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        """Instantiate a model from local directory or remote model repo. Note
        that when loading from remote, the model revision can be specified.

        Args:
            model_name_or_path(str): A model dir or a model id to be loaded
            revision(str, `optional`): The revision used when the model_name_or_path is
                a model id of the remote hub. default `master`.
            cfg_dict(Config, `optional`): An optional model config. If provided, it will replace
                the config read out of the `model_name_or_path`
            device(str, `optional`): The device to load the model.
            **kwargs:
                task(str, `optional`): The `Tasks` enumeration value to replace the task value
                read out of config in the `model_name_or_path`. This is useful when the model to be loaded is not
                equal to the model saved.
                For example, load a `backbone` into a `text-classification` model.
                Other kwargs will be directly fed into the `model` key, to replace the default configs.
        Returns:
            A model instance.

        Examples:
            >>> from modelscope.models import Model
            >>> Model.from_pretrained('damo/nlp_structbert_backbone_base_std', task='text-classification')
        """
        prefetched = kwargs.get('model_prefetched')
        if prefetched is not None:
            kwargs.pop('model_prefetched')

        if osp.exists(model_name_or_path):
            local_model_dir = model_name_or_path
        else:
            if prefetched is True:
                raise RuntimeError('Expecting model is pre-fetched locally, but is not found.')
            local_model_dir = snapshot_download(model_name_or_path, revision)
        logger.info(f'initialize model from {local_model_dir}')
        if cfg_dict is not None:
            cfg = cfg_dict
        else:
            cfg = Config.from_file(osp.join(local_model_dir, ModelFile.CONFIGURATION))

        model_cfg = cfg.model
        if hasattr(model_cfg, 'model_type') and not hasattr(model_cfg, 'type'):
            model_cfg.type = model_cfg.model_type

        for k, v in kwargs.items():
            model_cfg[k] = v
        if device is not None:
            model_cfg.device = device

        model = build_model(model_cfg, default_args=kwargs)
        model.model_dir = local_model_dir

        # dynamically add pipeline info to model for pipeline inference
        if hasattr(cfg, 'pipeline'):
            model.pipeline = cfg.pipeline
        else:
            model.pipeline = ConfigDict(type=cfg.task)

        # cfg['framework'] = 'pytorch'
        for key in ('cfg', 'config'):
            if not hasattr(model, key):
                setattr(model, key, cfg)

        model.name = model_name_or_path
        return model



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Optional, Set

from adaseq.metainfo import DataCollators

from .base import DATA_COLLATORS, DataCollatorWithPadding


@DATA_COLLATORS.register_module(module_name=DataCollators.pretraining_data_collator)
class PretrainingDataCollatorWithPadding(DataCollatorWithPadding):
    """Padding method for pretraining dataset."""

    def __init__(
        self, tokenizer, default_pad_id: int = 0, no_pad_fields: Optional[Set[str]] = None, **kwargs
    ) -> None:
        super().__init__(tokenizer, default_pad_id, no_pad_fields, **kwargs)
        self.keep_fields.add('prompt_type')



# Copyright (c) Alibaba, Inc. and its affiliates.
from os import path as osp
from typing import Dict, Union

import torch.nn as nn
from modelscope.utils.config import Config, ConfigDict
from modelscope.utils.registry import Registry, build_from_cfg

ENCODERS = Registry('encoders')


def build_encoder(cfg: ConfigDict, default_args: dict = None):
    """Build encoder from config dict

    Args:
        cfg (:obj:`ConfigDict`): config dict for encoder object
        default_args (dict): default initialization arguments

    Returns:
        encoder (:obj:`Encoder`): an encoder instance
    """
    return build_from_cfg(cfg, ENCODERS, group_key='default', default_args=default_args)


class Encoder(nn.Module):
    """
    The encoder base class for encoding embeddings to features.
    """

    def get_input_dim(self) -> int:
        """
        Get the input feature dim.
        """
        raise NotImplementedError

    def get_output_dim(self) -> int:
        """
        Get the output feature dim.
        """
        raise NotImplementedError

    @classmethod
    def from_config(cls, cfg_dict_or_path: Union[str, Dict, Config] = None, **kwargs):
        """Build encoder instance from config"""
        if isinstance(cfg_dict_or_path, str) and osp.isfile(cfg_dict_or_path):
            cfg = Config.from_file(cfg_dict_or_path).model.encoder
        elif isinstance(cfg_dict_or_path, (dict, Config)):
            cfg = cfg_dict_or_path
        else:
            cfg = {}

        if 'type' not in cfg:
            cfg['type'] = None

        if 'type' in kwargs:
            cfg['type'] = kwargs.pop('type')

        if cfg['type'] is not None and cfg['type'] in ENCODERS.modules['default']:
            return build_encoder(cfg, default_args=kwargs)
        else:
            raise ValueError('Unsupported encoder type: %s', cfg['type'])



# Copyright (c) Alibaba, Inc. and its affiliates.
"""
Functions and exceptions for checking that
AdaSeq and its models are configured correctly.
"""
from typing import Any, Tuple, Union


# Copyright (c) AI2 AllenNLP. Licensed under the Apache License, Version 2.0.
class ConfigurationError(Exception):
    """
    The exception raised by any Adaseq object when it's misconfigured
    (e.g. missing properties, invalid properties, unknown properties).
    """

    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
        return type(self), (self.message,)

    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def __str__(self):
        return self.message



# Copyright (c) Alibaba, Inc. and its affiliates.
from .base import DECODERS, Decoder, build_decoder
from .crf import CRF, CRFwithConstraints
from .mlm_head import OnlyMLMHead
from .pairwise_crf import PairwiseCRF
from .partial_crf import PartialCRF



# Copyright (c) Alibaba, Inc. and its affiliates.

import datasets
from datasets import Features, Value

from adaseq.data.constant import (
    NONE_REL_LABEL,
    OBJECT_START_TOKEN,
    PAD_LABEL,
    SUBJECT_START_TOKEN,
)

from .base import CustomDatasetBuilder


class RelationExtractionDatasetBuilderConfig(datasets.BuilderConfig):
    """Builder Config for Relation Extraction"""

    def __init__(self, data_dir=None, data_files=None, **corpus_config):
        super().__init__(data_dir=data_dir, data_files=data_files)
        self.corpus_config = corpus_config


class RelationExtractionDatasetBuilder(CustomDatasetBuilder):
    """Dataset Builder for Relation Extraction"""

    BUILDER_CONFIG_CLASS = RelationExtractionDatasetBuilderConfig

    def stub():  # noqa: D102
        pass

    @classmethod
    def parse_label(cls, data):  # noqa: D102
        return [data['label']]  # TODO fix

    def _info(self):
        info = datasets.DatasetInfo(
            features=Features(
                {
                    'id': Value('string'),
                    'tokens': [Value('string')],
                    'mask': [Value('bool')],
                    'so_head_mask': [Value('bool')],
                    'label': Value('string'),
                }
            )
        )
        return info

    def _generate_examples(self, filepath):
        corpus_config = self.config.corpus_config
        if corpus_config['data_type'] == 'conll':
            return self._load_conll_file(filepath, corpus_config)
        else:
            raise ValueError('Unknown corpus format type [%s]' % corpus_config['data_type'])

    @classmethod
    def load_data_file(cls, file_path, corpus_config):
        """load CoNLL format file."""
        if corpus_config['data_type'] == 'conll':
            return cls._load_conll_file(file_path, corpus_config)
        else:
            raise ValueError('Unknown corpus format type [%s]' % corpus_config['data_type'])

    @classmethod
    def _load_conll_file(cls, file_path, corpus_config):
        delimiter = corpus_config.get('delimiter', None)

        with open(file_path, encoding='utf-8') as f:
            guid = 0
            tokens = []
            labels = []
            for line in f:
                if line.startswith('-DOCSTART-') or line == '' or line == '\n':
                    if tokens:
                        mask = cls._labels_to_mask(labels)
                        so_head_mask = cls._create_so_head_mask(tokens)
                        label = cls._extract_rel_label(tokens, labels)
                        yield guid, {
                            'id': str(guid),
                            'tokens': tokens,
                            'label': label,
                            'mask': mask,
                            'so_head_mask': so_head_mask,
                        }
                        guid += 1
                        tokens = []
                        labels = []
                else:
                    splits = line.split(delimiter)
                    tokens.append(splits[0])
                    labels.append(splits[-1].rstrip())

            if tokens:
                mask = cls._labels_to_mask(labels)
                so_head_mask = cls._create_so_head_mask(tokens)
                label = cls._extract_rel_label(tokens, labels)
                yield guid, {
                    'id': str(guid),
                    'tokens': tokens,
                    'label': label,
                    'mask': mask,
                    'so_head_mask': so_head_mask,
                }

    @classmethod
    def _extract_rel_label(cls, tokens, labels):
        """
        example:
        token   label
        the     O
        <E>     /misc/misc/part_of
        Circle  B-A
        Undone  I-A
        </E>    O
        has     O
        been    O
        released        O
        """
        rel_label = NONE_REL_LABEL
        for token, label in zip(tokens, labels):
            if token in (SUBJECT_START_TOKEN, OBJECT_START_TOKEN):
                rel_label = label
                break
        return rel_label

    @classmethod
    def _create_so_head_mask(cls, tokens):
        mask = []
        for token in tokens:
            mask.append(token in (SUBJECT_START_TOKEN, OBJECT_START_TOKEN))
        return mask

    @classmethod
    def _labels_to_mask(cls, labels):
        mask = []
        for label in labels:
            mask.append(label != PAD_LABEL)
        return mask



# Copyright (c) Alibaba, Inc. and its affiliates.
from .multilabel_typing_preprocessor import (
    MultiLabelConcatTypingPreprocessor,
    MultiLabelSpanTypingPreprocessor,
)
from .nlp_preprocessor import NLPPreprocessor
from .pretraining_preprocessor import PretrainingPreprocessor
from .relation_extraction_preprocessor import RelationExtractionPreprocessor
from .sequence_labeling_preprocessor import SequenceLabelingPreprocessor
from .span_extraction_preprocessor import SpanExtracionPreprocessor
from .twostage_preprocessor import TwoStagePreprocessor



# Copyright (c) Alibaba, Inc. and its affiliates.
import datetime
from typing import Dict

from modelscope.utils.config import Config


def create_datetime_str() -> str:
    """Create a string indicating current time

    Create a string indicating current time in microsecond precision,
    for example, 221109144626.861616

    Returns:
        str: current time string
    """
    datetime_dt = datetime.datetime.today()
    datetime_str = datetime_dt.strftime('%y%m%d%H%M%S.%f')
    return datetime_str


def has_keys(_dict: Dict, *keys: str):
    """Check whether a nested dict has a key

    Args:
        _dict (Dict): a nested dict like object
        *keys (str): flattened key list

    Returns:
        bool: whether _dict has keys
    """
    if not _dict or not keys:
        return False

    sub_dict = _dict
    for key in keys:
        if isinstance(sub_dict, (dict, Config)) and key in sub_dict:
            sub_dict = sub_dict[key]
        else:
            return False
    return True



# Copyright (c) Alibaba, Inc. and its affiliates.
import argparse
from abc import ABC, abstractmethod


class Subcommand(ABC):
    """Abstract class for subcommands"""

    @classmethod
    @abstractmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> argparse.ArgumentParser:
        """Add arguments parser for subcommand"""
        pass



# Copyright (c) Alibaba, Inc. and its affiliates.
import os
import pickle
from typing import Optional

import numpy as np
import torch
from torch import nn
from tqdm import tqdm

from adaseq.metainfo import Decoders

from .base import DECODERS, Decoder


def make_glove_embed(tokens, glove_path, embed_dim=100):
    """Utils function to obtain glove embedding"""

    glove = {}
    vecs = []  # use to produce unk

    i = 0
    # load glove
    with open(os.path.join(glove_path), 'r', encoding='utf-8') as f:
        for line in tqdm(f.readlines()):
            i += 1

            split_line = line.split()
            word = split_line[0]
            embed_str = split_line[1:]
            try:
                # ignore error
                embed_float = [float(i) for i in embed_str]
            except Exception:
                continue

            if word not in glove:
                glove[word] = embed_float
                vecs.append(embed_float)

    unk = np.zeros(embed_dim)
    print('error lines: {}'.format(i - len(vecs)))

    print('loading glove to task vocab')
    # load glove to task vocab
    embed = []
    for i in tqdm(tokens):
        word = i.lower().split('_')
        sub_emb = []
        for sub in word:
            if sub in glove:
                sub_emb.append(glove[sub])
            else:
                sub_emb.append(unk)
        emb = np.mean(sub_emb, 0).tolist()
        embed.append(emb)

    final_embed = np.array(embed, dtype=np.float)
    return final_embed


def make_tencent_embed(tokens, tencent_path, embed_dim=200):
    """Utils function to obtain tencent zh embedding"""
    glove = {}
    import jieba

    i = 0
    # load glove
    with open(os.path.join(tencent_path), 'r', encoding='utf-8') as f:
        for line in tqdm(f.readlines()):
            i += 1
            if i == 1:
                continue

            split_line = line.split()
            word = split_line[0]
            embed_str = split_line[1:]
            try:
                # ignore error
                embed_float = [float(i) for i in embed_str]
            except Exception:
                continue

            if word not in glove:
                glove[word] = embed_float

    unk = np.zeros(embed_dim)

    print('loading tencent to task vocab')

    embed = []
    for i in tqdm(tokens):
        word = i.replace('(', '')
        word = word.replace(')', '')

        sub_emb = []
        if word in glove:
            sub_emb.append(glove[word])
        else:
            seg_list = jieba.cut(word, cut_all=True)
            for ww in seg_list:
                if ww in glove:
                    sub_emb.append(glove[ww])
                else:
                    sub_emb.append(unk)

        emb = np.mean(sub_emb, 0).tolist()

        embed.append(emb)

    final_embed = np.array(embed, dtype=np.float)

    return final_embed


def get_label_emb(
    labels,
    label_emb_type='glove',
    label_emb_dim=300,
    source_emb_file_path=None,
    target_emb_dir=None,
    target_emb_name='label.emb',
):
    """produce label embeddings"""
    if target_emb_dir is None:
        print('target emb is not given, returning random embedding')
        torch.randn(len(labels), label_emb_dim)

    os.makedirs(target_emb_dir, exist_ok=True)
    emb_save_path = os.path.join('{}/{}'.format(target_emb_dir, target_emb_name))
    if label_emb_type == 'tencent':
        if os.path.exists(emb_save_path):
            label_emb = torch.from_numpy(pickle.load(open(emb_save_path, 'rb'))).float()
        else:
            print(
                'no pre-processed emb file found, using source emb file to produce label embeddings'
            )
            assert source_emb_file_path is not None, 'require source emb file (tencent)'
            label_emb = make_tencent_embed(
                labels, tencent_path=source_emb_file_path, embed_dim=label_emb_dim
            )
            pickle.dump(label_emb, open(os.path.join(emb_save_path), 'wb'))
            label_emb = torch.from_numpy(label_emb).float()

    elif label_emb_type == 'glove':
        if os.path.exists(emb_save_path):
            label_emb = torch.from_numpy(pickle.load(open(emb_save_path, 'rb'))).float()
        else:
            print(
                'no pre-processed emb file found, using source emb file to produce label embeddings'
            )
            assert source_emb_file_path is not None, 'require source emb file (glove)'
            label_emb = make_glove_embed(
                labels, glove_path=source_emb_file_path, embed_dim=label_emb_dim
            )
            pickle.dump(label_emb, open(os.path.join(emb_save_path), 'wb'))
            label_emb = torch.from_numpy(label_emb).float()

    else:
        raise NotImplementedError

    return label_emb


class SimpleFeedForwardLayer(nn.Module):
    """2-layer feed forward"""

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        bias: bool = True,
        activation: Optional[nn.Module] = None,
        dropout_rate: float = 0,
    ):

        super(SimpleFeedForwardLayer, self).__init__()
        self.linear_projection1 = nn.Linear(input_dim, (input_dim + output_dim) // 2, bias=bias)
        self.linear_projection_mid1 = nn.Linear(
            (input_dim + output_dim) // 2, (input_dim + output_dim) // 2, bias=bias
        )
        self.linear_projection_mid2 = nn.Linear(
            (input_dim + output_dim) // 2, (input_dim + output_dim) // 2, bias=bias
        )

        self.linear_projection2 = nn.Linear((input_dim + output_dim) // 2, output_dim, bias=bias)
        self.activation = activation if activation else nn.Tanh()
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:  # noqa: D102

        inputs = self.activation(self.dropout(self.linear_projection1(inputs)))
        inputs = self.linear_projection2(inputs)
        return inputs


class LabelTransformLayerMLP(nn.Module):
    """MLP on label embeddings to parameterize potentials"""

    def __init__(self, emb_dim, dropout_rate):
        super(LabelTransformLayerMLP, self).__init__()
        self.emb_dim = emb_dim
        self.dropout_rate = dropout_rate

        act = nn.Tanh()

        self.label_emb_view = nn.ModuleList(
            [
                SimpleFeedForwardLayer(
                    input_dim=self.emb_dim,
                    output_dim=self.emb_dim,
                    bias=False,
                    activation=act,
                    dropout_rate=self.dropout_rate,
                )
                for i in range(4)
            ]
        )

    def forward(self, label_emb):  # noqa: D102

        return (
            self.label_emb_view[0](label_emb).unsqueeze(2),
            self.label_emb_view[1](label_emb).unsqueeze(2),
            self.label_emb_view[2](label_emb).unsqueeze(2),
            self.label_emb_view[3](label_emb).unsqueeze(2),
        )


@DECODERS.register_module(module_name=Decoders.pairwise_crf)
class PairwiseCRF(Decoder):
    """
    The module proposed in the EMNLP2022 Paper to model label correlations,
     require label embeddings and use mean-field variational inference.
    "Modeling Label Correlations for Ultra-Fine Entity Typing with Neural Pairwise Conditional Random Field"
    Inference also based on the paper "Regularized Frank-Wolfe for Dense CRFs"
    Args:
        labels: list of labels, for building the label embeddings.
        label_emb_dim: number of label embedding dimention.
        label_emb_type: glove (en) / tencent (zh).
        source_emb_file_path: path of label embed file, e.g. glove.6B.300.txt.
        target_emb_dir: dir to save label embedding.
        target_emb_name: name of saved label embedding.
        pairwise_factor: important args to balancing unary and pairwise potentials.
        two_potential: whether to use two pairwise potential.
        sign_trick: whether to use two pairwise potential.
        mfvi_step_size: step size in Regularized Frank-Wolfe, default 1.0.
        mfvi_scaler: temperature, default 1.0.
        mfvi_iteration: number of iterations for mean-field variational inference.
        return_logits: true if reture logits, otherwise return probabilities.
        label_dropout: label embedding dropout rate.
        **kwargs:
    """

    def __init__(
        self,
        labels,
        label_emb_dim: int = 300,
        label_emb_type: str = 'glove',
        source_emb_file_path: Optional[str] = None,
        target_emb_dir: Optional[str] = None,
        target_emb_name: str = 'label.emb',
        pairwise_factor: int = 20,
        two_potential: bool = True,
        sign_trick: bool = True,
        mfvi_step_size: float = 1.0,
        mfvi_scaler: float = 1.0,
        mfvi_iteration: int = 3,
        return_logits: bool = True,
        label_dropout: float = 0.0,
        **kwargs
    ):
        super(PairwiseCRF, self).__init__()

        label_emb = get_label_emb(
            labels,
            label_emb_type=label_emb_type,
            label_emb_dim=label_emb_dim,
            source_emb_file_path=source_emb_file_path,
            target_emb_dir=target_emb_dir,
            target_emb_name=target_emb_name,
        )

        self.n_label, self.emb_dim = label_emb.size()
        assert self.emb_dim == label_emb_dim
        self.two_potential = two_potential
        self.pairwise_factor = pairwise_factor
        self.mfvi_step_size = mfvi_step_size
        self.mfvi_scaler = mfvi_scaler
        self.mfvi_iteration = mfvi_iteration
        self.sign_trick = sign_trick
        self.return_logits = return_logits
        self.label_dropout = label_dropout

        self.label_emb = nn.Parameter(label_emb / self.pairwise_factor, requires_grad=True)

        self.domain = 2
        self.sigmoid = nn.Sigmoid()
        self.mfvi_iteration = mfvi_iteration
        self.is_cuda = torch.cuda.is_available()
        self.label_emb_transform = LabelTransformLayerMLP(self.emb_dim, self.label_dropout)
        self.label_view = None

    def forward(self, logits):
        """Mean-field variational Inference"""

        # B = 3

        B, L = logits.size()

        logits_zero = (
            torch.zeros(B, self.n_label).cuda() if self.is_cuda else torch.zeros(B, self.n_label)
        )  # B x L
        q_logits = torch.cat([logits_zero.unsqueeze(0), logits.unsqueeze(0)], 0)  # 2 x B x L
        q_logits = q_logits / self.mfvi_scaler
        l_1 = q_logits[1]
        q_dist = q_logits.softmax(0)  # 2 x B x L

        # 1st Term, Unary Energy Ψu(yi):
        #     measures the cost if the label assignment disagrees with the initial classifier.
        # obtain unary logits
        # obtain unary potential, the negative log likelihood of sigmoid logits
        E1, E2, E3, E4 = self.label_emb_transform(self.label_emb)  # L x D x 4

        if self.two_potential:
            label_view1 = torch.cat([E1, E1, E2, E2], 2)
            label_view2 = torch.cat([E1, E2, E1, E2], 2)
        else:
            label_view1 = torch.cat([E1, E2, E3, E4], 2)
            label_view2 = torch.cat([E1, E3, E2, E4], 2)

        for _ in range(self.mfvi_iteration):

            # 2nd Term, Pairwise Energy Ψp(yi, yj):
            #     measures the cost if two similar label (e.g. label semantics) take different labels.
            q_dist_j_0 = q_dist[0]
            q_dist_j_1 = q_dist[1]
            q_logits_i_0 = q_logits[0]
            q_logits_i_1 = q_logits[1]

            dists_mat = torch.cat(
                [
                    q_dist_j_0.unsqueeze(0),
                    q_dist_j_1.unsqueeze(0),
                    q_dist_j_0.unsqueeze(0),
                    q_dist_j_1.unsqueeze(0),
                ],
                0,
            )  # 4 x B x l
            tmp = torch.einsum('vbl,ldv->vbd', dists_mat, label_view1)
            update_logits = torch.einsum('vbd,ldv->vbl', tmp, label_view2)

            if self.sign_trick:
                l_0 = q_logits_i_0 + update_logits[0] - update_logits[1]
                l_1 = q_logits_i_1 - update_logits[2] + update_logits[3]
            else:
                l_0 = q_logits_i_0 + update_logits[0] + update_logits[1]
                l_1 = q_logits_i_1 + update_logits[2] + update_logits[3]

            new_q_dist = (
                torch.cat([l_0.unsqueeze(0), l_1.unsqueeze(0)], dim=0) / self.mfvi_scaler
            ).softmax(0)
            q_dist = (1 - self.mfvi_step_size) * q_dist + self.mfvi_step_size * new_q_dist

        if self.return_logits:
            return l_1 / self.mfvi_scaler
        else:
            return q_dist[1]  # positive probs B x L



# Copyright (c) Alibaba, Inc. and its affiliates.
import os

from modelscope.hub.snapshot_download import snapshot_download
from modelscope.metainfo import Models
from modelscope.utils.hub import get_model_type
from transformers import AutoTokenizer, BertTokenizer, BertTokenizerFast
from transformers.tokenization_utils import PreTrainedTokenizer


def build_tokenizer(
    model_name_or_path: str, use_fast: bool = True, is_word2vec: bool = False, **kwargs
) -> PreTrainedTokenizer:
    """build tokenizer from `transformers`."""
    tokenizer = BertTokenizerFast if use_fast else BertTokenizer

    if is_word2vec or 'nezha' in model_name_or_path:
        return tokenizer.from_pretrained(model_name_or_path, **kwargs)

    try:
        return AutoTokenizer.from_pretrained(model_name_or_path, use_fast=use_fast, **kwargs)

    except OSError:
        if not os.path.exists(model_name_or_path):
            model_name_or_path = snapshot_download(model_name_or_path, **kwargs)
        # code borrowed from modelscope/preprocessors/nlp/nlp_base.py
        model_type = get_model_type(model_name_or_path)

        if model_type in (Models.structbert, Models.gpt3, Models.palm, Models.plug):
            return tokenizer.from_pretrained(model_name_or_path, **kwargs)

        elif model_type == Models.veco:
            from transformers import XLMRobertaTokenizer, XLMRobertaTokenizerFast

            tokenizer = XLMRobertaTokenizerFast if use_fast else XLMRobertaTokenizer

            return tokenizer.from_pretrained(model_name_or_path, **kwargs)

        elif model_type == Models.deberta_v2:
            from modelscope.models.nlp.deberta_v2 import (
                DebertaV2Tokenizer,
                DebertaV2TokenizerFast,
            )

            tokenizer = DebertaV2TokenizerFast if use_fast else DebertaV2Tokenizer
            return tokenizer.from_pretrained(model_name_or_path, **kwargs)

        else:
            raise ValueError('Unsupported tokenizer')



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Callable, Dict, List, Optional, Set, Union

from modelscope.utils.config import ConfigDict
from modelscope.utils.registry import Registry, build_from_cfg, default_group
from transformers import PreTrainedTokenizerBase

from adaseq.data.batch import DataBatch
from adaseq.metainfo import DataCollators

DATA_COLLATORS = Registry('data_collators')


def build_data_collator(
    tokenizer: PreTrainedTokenizerBase, cfg: ConfigDict, default_args: Optional[dict] = None
):
    """build data collator from config."""
    if default_args is None:
        default_args = {}
    default_args['tokenizer'] = tokenizer
    return build_from_cfg(cfg, DATA_COLLATORS, group_key=default_group, default_args=default_args)


@DATA_COLLATORS.register_module(module_name=DataCollators.data_collator_with_padding)
class DataCollatorWithPadding:
    """
    A `DataCollator` support padding some fields to same length.
    Support padding encoder related fields: input_ids, token_type_ids, mask,
    and padding other fields with `default_pad_id`.
    `no_pad_fields` will be skipped.
    """

    tokenizer: PreTrainedTokenizerBase

    def __init__(
        self, tokenizer, default_pad_id: int = 0, no_pad_fields: Optional[Set[str]] = None, **kwargs
    ) -> None:
        self.tokenizer = tokenizer
        self.default_pad_id = default_pad_id
        self.keep_fields: Set[str] = {'tokens', 'origin_tokens', 'meta'}
        self.no_pad_fields = no_pad_fields or set()

    @staticmethod
    def _get_pad_func(padding_side: str) -> Callable:
        if padding_side == 'right':

            def _pad(array, size: int, pad_value):
                return array + [pad_value] * size

        elif padding_side == 'left':

            def _pad(array, size: int, pad_value):
                return [pad_value] * size + array

        else:
            raise ValueError('Invalid padding strategy:' + str(padding_side))
        return _pad

    def padding_token(self, batch: Dict[str, Any], padding_side: str) -> Dict[str, Any]:
        """pad token related fields (hf.transformers style)"""
        _pad = self._get_pad_func(padding_side)

        batch_size = len(batch['meta'])
        for field in [f for f in batch.keys() if f.endswith('tokens')]:
            sub_field_pair = [
                ('input_ids', self.tokenizer.pad_token_id),
                ('attention_mask', False),
                ('mask', False),
            ]
            if 'token_type_ids' in batch[field][0]:
                sub_field_pair.append(('token_type_ids', self.tokenizer.pad_token_type_id))
            if 'offsets' in batch[field][0]:
                sub_field_pair.append(('offsets', (0, 0)))

            sub = 'has_special_tokens'
            try:
                padded_tokens = {sub: [batch[field][i][sub] for i in range(batch_size)]}
            except KeyError:
                padded_tokens = {}

            for sub, pad_value in sub_field_pair:
                if sub not in batch[field][0]:
                    continue
                padded_field = list()
                max_length = max(len(i[sub]) for i in batch[field])
                for i in range(batch_size):
                    difference = max_length - len(batch[field][i][sub])
                    if difference > 0:
                        padded_field.append(_pad(batch[field][i][sub], difference, pad_value))
                    else:
                        padded_field.append(batch[field][i][sub])
                padded_tokens[sub] = padded_field
            batch[field] = padded_tokens

        field = 'origin_mask'
        if field in batch:
            max_length = max(len(i) for i in batch[field])
            for i in range(batch_size):
                difference = max_length - len(batch[field][i])
                if difference > 0:
                    batch[field][i] = _pad(batch[field][i], difference, False)

        return batch

    def padding(
        self,
        batch: Dict[str, Any],
        padding_side: str,
        fields: Optional[Union[Set[str], str]] = None,
        pad_id: Optional[int] = None,
    ) -> Dict[str, Any]:
        """pad other fields."""
        pad_id = self.default_pad_id if pad_id is None else pad_id
        _pad = self._get_pad_func(padding_side)

        if fields is None:
            fields = set(batch.keys())
        elif isinstance(fields, str):
            fields = {fields}
        fields -= self.keep_fields.union(self.no_pad_fields)

        for field in fields:
            if not isinstance(batch[field][0], list):
                continue
            max_length = max(len(i) for i in batch[field])
            for i in range(len(batch[field])):
                difference = max_length - len(batch[field][i])
                if difference > 0:
                    batch[field][i] = _pad(batch[field][i], difference, pad_id)
        return batch

    def __call__(self, instances: List[Dict[str, Any]]) -> DataBatch:
        """pad list of instances to batch"""
        batch = {key: [one[key] for one in instances] for key in instances[0].keys()}
        padding_side = self.tokenizer.padding_side
        padded_batch = self.padding_token(batch, padding_side)
        padded_batch = self.padding(padded_batch, padding_side=padding_side)
        batch = DataBatch(padded_batch, self.keep_fields)
        return batch



# Copyright (c) Alibaba, Inc. and its affiliates.
# Some functions are borrowed from AllenNLP.
# https://github.com/allenai/allennlp/blob/HEAD/allennlp/nn/util.py
# Copyright (c) AI2 AllenNLP. Licensed under the Apache License, Version 2.0.
from typing import Dict, Optional

import torch


def get_tokens_mask(tokens: Dict[str, torch.Tensor], size: int) -> torch.BoolTensor:
    """
    Return (sub-)token mask that removed special tokens.
    """
    if 'mask' in tokens:
        mask = tokens['mask']
    else:
        mask = tokens['attention_mask']
    if mask.size(1) > size:
        # remove top num added special tokens positions
        mask = mask[:, mask.size(1) - size :]
    return mask  # type: ignore


def tiny_value_of_dtype(dtype: torch.dtype):
    """
    Returns a moderately tiny value for a given PyTorch data type that is used to avoid numerical
    issues such as division by zero.
    This is different from `info_value_of_dtype(dtype).tiny` because it causes some NaN bugs.
    Only supports floating point dtypes.
    """
    if not dtype.is_floating_point:
        raise TypeError('Only supports floating point dtypes.')
    if dtype == torch.float or dtype == torch.double:
        return 1e-13
    elif dtype == torch.half:
        return 1e-4
    else:
        raise TypeError('Does not support dtype ' + str(dtype))


def batched_span_select(target: torch.Tensor, spans: torch.LongTensor) -> torch.Tensor:
    """
    The given `spans` of size `(batch_size, num_spans, 2)` indexes into the sequence
    dimension (dimension 2) of the target, which has size `(batch_size, sequence_length,
    embedding_size)`.

    This function returns segmented spans in the target with respect to the provided span indices.

    # Parameters

    target : `torch.Tensor`, required.
        A 3 dimensional tensor of shape (batch_size, sequence_length, embedding_size).
        This is the tensor to be indexed.
    indices : `torch.LongTensor`
        A 3 dimensional tensor of shape (batch_size, num_spans, 2) representing start and end
        indices (both inclusive) into the `sequence_length` dimension of the `target` tensor.

    # Returns

    span_embeddings : `torch.Tensor`
        A tensor with shape (batch_size, num_spans, max_batch_span_width, embedding_size]
        representing the embedded spans extracted from the batch flattened target tensor.
    span_mask: `torch.BoolTensor`
        A tensor with shape (batch_size, num_spans, max_batch_span_width) representing the mask on
        the returned span embeddings.
    """
    # both of shape (batch_size, num_spans, 1)
    span_starts, span_ends = spans.split(1, dim=-1)

    # shape (batch_size, num_spans, 1)
    # These span widths are off by 1, because the span ends are `inclusive`.
    span_widths = span_ends - span_starts

    # We need to know the maximum span width so we can
    # generate indices to extract the spans from the sequence tensor.
    # These indices will then get masked below, such that if the length
    # of a given span is smaller than the max, the rest of the values
    # are masked.
    max_batch_span_width = span_widths.max().item() + 1

    # Shape: (1, 1, max_batch_span_width)
    max_span_range_indices = get_range_vector(max_batch_span_width, get_device_of(target)).view(
        1, 1, -1
    )
    # Shape: (batch_size, num_spans, max_batch_span_width)
    # This is a broadcasted comparison - for each span we are considering,
    # we are creating a range vector of size max_span_width, but masking values
    # which are greater than the actual length of the span.
    #
    # We're using <= here (and for the mask below) because the span ends are
    # inclusive, so we want to include indices which are equal to span_widths rather
    # than using it as a non-inclusive upper bound.
    span_mask = max_span_range_indices <= span_widths
    raw_span_indices = span_starts + max_span_range_indices
    # We also don't want to include span indices which greater than the sequence_length,
    # which happens because some spans near the end of the sequence
    # have a start index + max_batch_span_width > sequence_length, so we add this to the mask here.
    span_mask = span_mask & (raw_span_indices < target.size(1)) & (0 <= raw_span_indices)
    span_indices = raw_span_indices * span_mask

    # Shape: (batch_size, num_spans, max_batch_span_width, embedding_dim)
    span_embeddings = batched_index_select(target, span_indices)

    return span_embeddings, span_mask


def batched_index_select(
    target: torch.Tensor,
    indices: torch.LongTensor,
    flattened_indices: Optional[torch.LongTensor] = None,
) -> torch.Tensor:
    """
    The given `indices` of size `(batch_size, d_1, ..., d_n)` indexes into the sequence
    dimension (dimension 2) of the target, which has size `(batch_size, sequence_length,
    embedding_size)`.

    This function returns selected values in the target with respect to the provided indices, which
    have size `(batch_size, d_1, ..., d_n, embedding_size)`. This can use the optionally
    precomputed `flattened_indices` with size `(batch_size * d_1 * ... * d_n)` if given.

    An example use case of this function is looking up the start and end indices of spans in a
    sequence tensor. This is used in the
    [CoreferenceResolver](https://docs.allennlp.org/models/main/models/coref/models/coref/)
    model to select contextual word representations corresponding to the start and end indices of
    mentions.

    The key reason this can't be done with basic torch functions is that we want to be able to use look-up
    tensors with an arbitrary number of dimensions (for example, in the coref model, we don't know
    a-priori how many spans we are looking up).

    # Parameters

    target : `torch.Tensor`, required.
        A 3 dimensional tensor of shape (batch_size, sequence_length, embedding_size).
        This is the tensor to be indexed.
    indices : `torch.LongTensor`
        A tensor of shape (batch_size, ...), where each element is an index into the
        `sequence_length` dimension of the `target` tensor.
    flattened_indices : `Optional[torch.Tensor]`, optional (default = `None`)
        An optional tensor representing the result of calling `flatten_and_batch_shift_indices`
        on `indices`. This is helpful in the case that the indices can be flattened once and
        cached for many batch lookups.

    # Returns

    selected_targets : `torch.Tensor`
        A tensor with shape [indices.size(), target.size(-1)] representing the embedded indices
        extracted from the batch flattened target tensor.
    """
    if flattened_indices is None:
        # Shape: (batch_size * d_1 * ... * d_n)
        flattened_indices = flatten_and_batch_shift_indices(indices, target.size(1))

    # Shape: (batch_size * sequence_length, embedding_size)
    flattened_target = target.view(-1, target.size(-1))

    # Shape: (batch_size * d_1 * ... * d_n, embedding_size)
    flattened_selected = flattened_target.index_select(0, flattened_indices)
    selected_shape = list(indices.size()) + [target.size(-1)]
    # Shape: (batch_size, d_1, ..., d_n, embedding_size)
    selected_targets = flattened_selected.view(*selected_shape)
    return selected_targets


def flatten_and_batch_shift_indices(indices: torch.Tensor, sequence_length: int) -> torch.Tensor:
    """
    This is a subroutine for [`batched_index_select`](./util.md#batched_index_select).
    The given `indices` of size `(batch_size, d_1, ..., d_n)` indexes into dimension 2 of a
    target tensor, which has size `(batch_size, sequence_length, embedding_size)`. This
    function returns a vector that correctly indexes into the flattened target. The sequence
    length of the target must be provided to compute the appropriate offsets.

    ```python
        indices = torch.ones([2,3], dtype=torch.long)
        # Sequence length of the target tensor.
        sequence_length = 10
        shifted_indices = flatten_and_batch_shift_indices(indices, sequence_length)
        # Indices into the second element in the batch are correctly shifted
        # to take into account that the target tensor will be flattened before
        # the indices are applied.
        assert shifted_indices == [1, 1, 1, 11, 11, 11]
    ```

    # Parameters

    indices : `torch.LongTensor`, required.
    sequence_length : `int`, required.
        The length of the sequence the indices index into.
        This must be the second dimension of the tensor.

    # Returns

    offset_indices : `torch.LongTensor`
    """
    # Shape: (batch_size)
    if torch.max(indices) >= sequence_length or torch.min(indices) < 0:
        raise ValueError(f'All elements in indices should be in range (0, {sequence_length - 1})')
    offsets = get_range_vector(indices.size(0), get_device_of(indices)) * sequence_length
    for _ in range(len(indices.size()) - 1):
        offsets = offsets.unsqueeze(1)

    # Shape: (batch_size, d_1, ..., d_n)
    offset_indices = indices + offsets

    # Shape: (batch_size * d_1 * ... * d_n)
    offset_indices = offset_indices.view(-1)
    return offset_indices


def get_range_vector(size: int, device: int) -> torch.Tensor:
    """
    Returns a range vector with the desired size, starting at 0. The CUDA implementation
    is meant to avoid copy data from CPU to GPU.
    """
    if device > -1:
        return torch.cuda.LongTensor(size, device=device).fill_(1).cumsum(0) - 1
    else:
        return torch.arange(0, size, dtype=torch.long)


def get_device_of(tensor: torch.Tensor) -> int:
    """
    Returns the device of the tensor.
    """
    if not tensor.is_cuda:
        return -1
    else:
        return tensor.get_device()



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import List, Optional

import torch
import torch.nn as nn

from adaseq.metainfo import Decoders

from .base import DECODERS, Decoder


@DECODERS.register_module(module_name=Decoders.crf)
class CRF(Decoder):
    """Conditional random field.
    This module implements a conditional random field [LMP01]_. The forward computation
    of this class computes the log likelihood of the given sequence of tags and
    emission score tensor. This class also has `~CRF.decode` method which finds
    the best tag sequence given an emission score tensor using `Viterbi algorithm`_.
    Args:
        num_tags: Number of tags.
        batch_first: Whether the first dimension corresponds to the size of a minibatch.
    Attributes:
        start_transitions (`~torch.nn.Parameter`): Start transition score tensor of size
            ``(num_tags,)``.
        end_transitions (`~torch.nn.Parameter`): End transition score tensor of size
            ``(num_tags,)``.
        transitions (`~torch.nn.Parameter`): Transition score tensor of size
            ``(num_tags, num_tags)``.
    .. [LMP01] Lafferty, J., McCallum, A., Pereira, F. (2001).
       "Conditional random fields: Probabilistic models for segmenting and
       labeling sequence data". *Proc. 18th International Conf. on Machine
       Learning*. Morgan Kaufmann. pp. 282–289.
    .. _Viterbi algorithm: https://en.wikipedia.org/wiki/Viterbi_algorithm
    """

    def __init__(self, num_tags: int, batch_first: bool = False) -> None:
        if num_tags <= 0:
            raise ValueError(f'invalid number of tags: {num_tags}')
        super().__init__()
        self.num_tags = num_tags
        self.batch_first = batch_first
        self.start_transitions = nn.Parameter(torch.empty(num_tags))
        self.end_transitions = nn.Parameter(torch.empty(num_tags))
        self.transitions = nn.Parameter(torch.empty(num_tags, num_tags))

        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Initialize the transition parameters.
        The parameters will be initialized randomly from a uniform distribution
        between -0.1 and 0.1.
        """
        nn.init.uniform_(self.start_transitions, -0.1, 0.1)
        nn.init.uniform_(self.end_transitions, -0.1, 0.1)
        nn.init.uniform_(self.transitions, -0.1, 0.1)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(num_tags={self.num_tags})'

    def forward(
        self,
        emissions: torch.Tensor,
        tags: torch.LongTensor,
        mask: Optional[torch.ByteTensor] = None,
        reduction: str = 'mean',
    ) -> torch.Tensor:
        """Compute the conditional log likelihood of a sequence of tags given emission scores.
        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length, num_tags)`` otherwise.
            tags (`~torch.LongTensor`): Sequence of tags tensor of size
                ``(seq_length, batch_size)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length)`` otherwise.
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
                if ``batch_first`` is ``False``, ``(batch_size, seq_length)`` otherwise.
            reduction: Specifies  the reduction to apply to the output:
                ``none|sum|mean|token_mean``. ``none``: no reduction will be applied.
                ``sum``: the output will be summed over batches. ``mean``: the output will be
                averaged over batches. ``token_mean``: the output will be averaged over tokens.
        Returns:
            `~torch.Tensor`: The log likelihood. This will have size ``(batch_size,)`` if
            reduction is ``none``, ``()`` otherwise.
        """
        if reduction not in ('none', 'sum', 'mean', 'token_mean'):
            raise ValueError(f'invalid reduction: {reduction}')
        if mask is None:
            mask = torch.ones_like(tags, dtype=torch.uint8, device=tags.device)
        if mask.dtype != torch.uint8:
            mask = mask.byte()
        self._validate(emissions, tags=tags, mask=mask)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            tags = tags.transpose(0, 1)
            mask = mask.transpose(0, 1)

        # shape: (batch_size,)
        numerator = self._compute_score(emissions, tags, mask)
        # shape: (batch_size,)
        denominator = self._compute_normalizer(emissions, mask)
        # shape: (batch_size,)
        llh = numerator - denominator

        if reduction == 'none':
            return llh
        if reduction == 'sum':
            return llh.sum()
        if reduction == 'mean':
            return llh.mean()
        return llh.sum() / mask.float().sum()

    def decode(
        self,
        emissions: torch.Tensor,
        mask: Optional[torch.ByteTensor] = None,
        nbest: Optional[int] = None,
        pad_tag: Optional[int] = None,
    ) -> List[List[List[int]]]:
        """Find the most likely tag sequence using Viterbi algorithm.
        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length, num_tags)`` otherwise.
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
                if ``batch_first`` is ``False``, ``(batch_size, seq_length)`` otherwise.
            nbest (`int`): Number of most probable paths for each sequence
            pad_tag (`int`): Tag at padded positions. Often input varies in length and
                the length will be padded to the maximum length in the batch. Tags at
                the padded positions will be assigned with a padding tag, i.e. `pad_tag`
        Returns:
            A PyTorch tensor of the best tag sequence for each batch of shape
            (nbest, batch_size, seq_length)
        """
        if nbest is None:
            nbest = 1
        if mask is None:
            mask = torch.ones(emissions.shape[:2], dtype=torch.uint8, device=emissions.device)
        if mask.dtype != torch.uint8:
            mask = mask.byte()
        self._validate(emissions, mask=mask)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            mask = mask.transpose(0, 1)

        if nbest == 1:
            return self._viterbi_decode(emissions, mask, pad_tag).unsqueeze(0)
        return self._viterbi_decode_nbest(emissions, mask, nbest, pad_tag)

    def _validate(
        self,
        emissions: torch.Tensor,
        tags: Optional[torch.LongTensor] = None,
        mask: Optional[torch.ByteTensor] = None,
    ) -> None:
        if emissions.dim() != 3:
            raise ValueError(f'emissions must have dimension of 3, got {emissions.dim()}')
        if emissions.size(2) != self.num_tags:
            raise ValueError(
                f'expected last dimension of emissions is {self.num_tags}, '
                f'got {emissions.size(2)}'
            )

        if tags is not None:
            if emissions.shape[:2] != tags.shape:
                raise ValueError(
                    'the first two dimensions of emissions and tags must match, '
                    f'got {tuple(emissions.shape[:2])} and {tuple(tags.shape)}'
                )

        if mask is not None:
            if emissions.shape[:2] != mask.shape:
                raise ValueError(
                    'the first two dimensions of emissions and mask must match, '
                    f'got {tuple(emissions.shape[:2])} and {tuple(mask.shape)}'
                )
            no_empty_seq = not self.batch_first and mask[0].all()
            no_empty_seq_bf = self.batch_first and mask[:, 0].all()
            if not no_empty_seq and not no_empty_seq_bf:
                raise ValueError('mask of the first timestep must all be on')

    def _compute_score(
        self, emissions: torch.Tensor, tags: torch.LongTensor, mask: torch.ByteTensor
    ) -> torch.Tensor:
        # emissions: (seq_length, batch_size, num_tags)
        # tags: (seq_length, batch_size)
        # mask: (seq_length, batch_size)
        seq_length, batch_size = tags.shape
        mask = mask.float()

        # Start transition score and first emission
        # shape: (batch_size,)
        score = self.start_transitions[tags[0]]
        score += emissions[0, torch.arange(batch_size), tags[0]]

        for i in range(1, seq_length):
            # Transition score to next tag, only added if next timestep is valid (mask == 1)
            # shape: (batch_size,)
            score += self.transitions[tags[i - 1], tags[i]] * mask[i]

            # Emission score for next tag, only added if next timestep is valid (mask == 1)
            # shape: (batch_size,)
            score += emissions[i, torch.arange(batch_size), tags[i]] * mask[i]

        # End transition score
        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1
        # shape: (batch_size,)
        last_tags = tags[seq_ends, torch.arange(batch_size)]
        # shape: (batch_size,)
        score += self.end_transitions[last_tags]

        return score

    def _forward_backward_algorithm(
        self,
        emissions: torch.Tensor,
        mask: torch.ByteTensor,
        mode: str = 'forward',
    ) -> torch.Tensor:
        """
        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)``
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
            mode (`str`): Specifies the calculation mode: ``partition|forward|backward``.
        Returns:
            A PyTorch tensor of the shape
            ``(batch_size, num_tags)`` for `partition`,
            ``(seq_length, batch_size, num_tags)`` for `forward`,
            the difference between `partition` and `forward` is whether add the `end_transtions` or not.
            ``(seq_length, batch_size, num_tags)`` for `backward`.

        """
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)

        if mode in ('forward', 'partition'):
            start_transitions = self.start_transitions
            end_transitions = self.end_transitions
            current_emissions = emissions
            transitions = self.transitions
        elif mode == 'backward':
            start_transitions = self.end_transitions
            end_transitions = self.start_transitions
            current_emissions = torch.zeros_like(emissions)
            batch_size = emissions.shape[1]
            for batch_idx in range(batch_size):
                length = mask[:, batch_idx].sum()
                current_emissions[:length, batch_idx, :] = emissions[:length, batch_idx, :].flip(
                    [0]
                )
            transitions = self.transitions.transpose(0, 1)
        else:
            raise NotImplementedError

        seq_length = current_emissions.size(0)
        scores = torch.zeros_like(current_emissions)

        # Start transition score and first emission; score has size of
        # (batch_size, num_tags) where for each batch, the j-th column stores
        # the score that the first timestep has tag j
        # shape: (batch_size, num_tags)
        score = start_transitions + current_emissions[0]

        scores[0, :, :] = score

        for i in range(1, seq_length):
            # Broadcast score for every possible next tag
            # shape: (batch_size, num_tags, 1)
            broadcast_score = score.unsqueeze(2)

            # Broadcast emission score for every possible current tag
            # shape: (batch_size, 1, num_tags)
            broadcast_emissions = current_emissions[i].unsqueeze(1)

            # Compute the score tensor of size (batch_size, num_tags, num_tags) where
            # for each sample, entry at row i and column j stores the sum of scores of all
            # possible tag sequences so far that end with transitioning from tag i to tag j
            # and emitting
            # shape: (batch_size, num_tags, num_tags)
            next_score = broadcast_score + transitions + broadcast_emissions

            # Sum over all possible current tags, but we're in score space, so a sum
            # becomes a log-sum-exp: for each sample, entry i stores the sum of scores of
            # all possible tag sequences so far, that end in tag i
            # shape: (batch_size, num_tags)
            next_score = torch.logsumexp(next_score, dim=1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # shape: (batch_size, num_tags)
            score = torch.where(mask[i].unsqueeze(1), next_score, score)
            scores[i, :, :] = score

        # End transition score
        # shape: (batch_size, num_tags)
        if mode == 'partition':
            score += end_transitions
        scores[seq_length - 1, :, :] = score

        if mode == 'backward':
            batch_size = emissions.shape[1]
            for batch_idx in range(batch_size):
                length = mask[:, batch_idx].sum()
                scores[:length, batch_idx, :] = (
                    scores[:length, batch_idx, :] - current_emissions[:length, batch_idx, :]
                ).flip([0])

        return scores

    def compute_posterior(self, emissions: torch.Tensor, mask: torch.ByteTensor) -> torch.Tensor:
        """Compute posterior probability distribution from emission logits

        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)``
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``

        Returns:
            A PyTorch tensor of the shape
            ``(seq_length, batch_size, num_tags)``
        """
        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            mask = mask.transpose(0, 1)

        fw_scores = self._forward_backward_algorithm(emissions, mask, mode='forward')
        bw_scores = self._forward_backward_algorithm(emissions, mask, mode='backward')
        partition = self._compute_normalizer(emissions, mask)
        log_posterior = fw_scores + bw_scores - partition.view(1, -1, 1)

        if self.batch_first:
            log_posterior = log_posterior.transpose(0, 1)
        return log_posterior

    def _compute_normalizer(self, emissions: torch.Tensor, mask: torch.ByteTensor) -> torch.Tensor:
        fw_scores = self._forward_backward_algorithm(emissions, mask, mode='partition')
        return torch.logsumexp(fw_scores[-1, :, :], dim=1)

    def _viterbi_decode(
        self, emissions: torch.FloatTensor, mask: torch.ByteTensor, pad_tag: Optional[int] = None
    ) -> List[List[int]]:
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        # return: (batch_size, seq_length)
        if pad_tag is None:
            pad_tag = 0

        device = emissions.device
        seq_length, batch_size = mask.shape

        # Start transition and first emission
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]
        history_idx = torch.zeros(
            (seq_length, batch_size, self.num_tags), dtype=torch.long, device=device
        )
        oor_idx = torch.zeros((batch_size, self.num_tags), dtype=torch.long, device=device)
        oor_tag = torch.full((seq_length, batch_size), pad_tag, dtype=torch.long, device=device)

        # - score is a tensor of size (batch_size, num_tags) where for every batch,
        #   value at column j stores the score of the best tag sequence so far that ends
        #   with tag j
        # - history_idx saves where the best tags candidate transitioned from; this is used
        #   when we trace back the best tag sequence
        # - oor_idx saves the best tags candidate transitioned from at the positions
        #   where mask is 0, i.e. out of range (oor)

        # Viterbi algorithm recursive case: we compute the score of the best tag sequence
        score = viterbi_decode_inner_loop1(
            score, history_idx, emissions, self.transitions, mask, oor_idx
        )
        # End transition score
        # shape: (batch_size, num_tags)
        end_score = score + self.end_transitions
        _, end_tag = end_score.max(dim=1)

        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1

        # insert the best tag at each sequence end (last position with mask == 1)
        history_idx = history_idx.transpose(1, 0).contiguous()
        history_idx.scatter_(
            1,
            seq_ends.view(-1, 1, 1).expand(-1, 1, self.num_tags),
            end_tag.view(-1, 1, 1).expand(-1, 1, self.num_tags),
        )
        history_idx = history_idx.transpose(1, 0).contiguous()

        # The most probable path for each sequence
        best_tags_arr = torch.zeros((seq_length, batch_size), dtype=torch.long, device=device)
        best_tags = torch.zeros(batch_size, 1, dtype=torch.long, device=device)

        best_tags_arr = viterbi_decode_inner_loop2(mask, history_idx, best_tags, best_tags_arr)
        return torch.where(mask, best_tags_arr, oor_tag).transpose(0, 1)

    def _viterbi_decode_nbest(
        self,
        emissions: torch.FloatTensor,
        mask: torch.ByteTensor,
        nbest: int,
        pad_tag: Optional[int] = None,
    ) -> List[List[List[int]]]:
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        # return: (nbest, batch_size, seq_length)
        if pad_tag is None:
            pad_tag = 0

        device = emissions.device
        seq_length, batch_size = mask.shape

        # Start transition and first emission
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]
        history_idx = torch.zeros(
            (seq_length, batch_size, self.num_tags, nbest), dtype=torch.long, device=device
        )
        oor_idx = torch.zeros((batch_size, self.num_tags, nbest), dtype=torch.long, device=device)
        oor_tag = torch.full(
            (seq_length, batch_size, nbest), pad_tag, dtype=torch.long, device=device
        )

        # + score is a tensor of size (batch_size, num_tags) where for every batch,
        #   value at column j stores the score of the best tag sequence so far that ends
        #   with tag j
        # + history_idx saves where the best tags candidate transitioned from; this is used
        #   when we trace back the best tag sequence
        # - oor_idx saves the best tags candidate transitioned from at the positions
        #   where mask is 0, i.e. out of range (oor)

        # Viterbi algorithm recursive case: we compute the score of the best tag sequence
        # for every possible next tag
        for i in range(1, seq_length):
            if i == 1:
                broadcast_score = score.unsqueeze(-1)
                broadcast_emission = emissions[i].unsqueeze(1)
                # shape: (batch_size, num_tags, num_tags)
                next_score = broadcast_score + self.transitions + broadcast_emission
            else:
                broadcast_score = score.unsqueeze(-1)
                broadcast_emission = emissions[i].unsqueeze(1).unsqueeze(2)
                # shape: (batch_size, num_tags, nbest, num_tags)
                next_score = broadcast_score + self.transitions.unsqueeze(1) + broadcast_emission

            # Find the top `nbest` maximum score over all possible current tag
            # shape: (batch_size, nbest, num_tags)
            next_score, indices = next_score.view(batch_size, -1, self.num_tags).topk(nbest, dim=1)

            if i == 1:
                score = score.unsqueeze(-1).expand(-1, -1, nbest)
                indices = indices * nbest

            # convert to shape: (batch_size, num_tags, nbest)
            next_score = next_score.transpose(2, 1)
            indices = indices.transpose(2, 1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # and save the index that produces the next score
            # shape: (batch_size, num_tags, nbest)
            score = torch.where(mask[i].unsqueeze(-1).unsqueeze(-1), next_score, score)
            indices = torch.where(mask[i].unsqueeze(-1).unsqueeze(-1), indices, oor_idx)
            history_idx[i - 1] = indices

        # End transition score shape: (batch_size, num_tags, nbest)
        end_score = score + self.end_transitions.unsqueeze(-1)
        _, end_tag = end_score.view(batch_size, -1).topk(nbest, dim=1)

        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1

        # insert the best tag at each sequence end (last position with mask == 1)
        history_idx = history_idx.transpose(1, 0).contiguous()
        history_idx.scatter_(
            1,
            seq_ends.view(-1, 1, 1, 1).expand(-1, 1, self.num_tags, nbest),
            end_tag.view(-1, 1, 1, nbest).expand(-1, 1, self.num_tags, nbest),
        )
        history_idx = history_idx.transpose(1, 0).contiguous()

        # The most probable path for each sequence
        best_tags_arr = torch.zeros(
            (seq_length, batch_size, nbest), dtype=torch.long, device=device
        )
        best_tags = (
            torch.arange(nbest, dtype=torch.long, device=device).view(1, -1).expand(batch_size, -1)
        )
        for idx in range(seq_length - 1, -1, -1):
            best_tags = torch.gather(history_idx[idx].view(batch_size, -1), 1, best_tags)
            best_tags_arr[idx] = best_tags.data.view(batch_size, -1) // nbest

        return torch.where(mask.unsqueeze(-1), best_tags_arr, oor_tag).permute(2, 1, 0)


@torch.jit.script
def viterbi_decode_inner_loop1(score, history_idx, emissions, transitions, mask, oor_idx):  # noqa
    # for every possible next tag

    seq_length, batch_size = mask.shape
    for i in range(1, seq_length):
        # Broadcast viterbi score for every possible next tag
        # shape: (batch_size, num_tags, 1)
        broadcast_score = score.unsqueeze(2)

        # Broadcast emission score for every possible current tag
        # shape: (batch_size, 1, num_tags)
        broadcast_emission = emissions[i].unsqueeze(1)

        # Compute the score tensor of size (batch_size, num_tags, num_tags) where
        # for each sample, entry at row i and column j stores the score of the best
        # tag sequence so far that ends with transitioning from tag i to tag j and emitting
        # shape: (batch_size, num_tags, num_tags)
        next_score = broadcast_score + transitions + broadcast_emission

        # Find the maximum score over all possible current tag
        # shape: (batch_size, num_tags)
        next_score, indices = next_score.max(dim=1)

        # Set score to the next score if this timestep is valid (mask == 1)
        # and save the index that produces the next score
        # shape: (batch_size, num_tags)
        score = torch.where(mask[i].unsqueeze(-1), next_score, score)
        indices = torch.where(mask[i].unsqueeze(-1), indices, oor_idx)
        history_idx[i - 1] = indices
    return score


@torch.jit.script
def viterbi_decode_inner_loop2(mask, history_idx, best_tags, best_tags_arr):  # noqa
    seq_length, batch_size = mask.shape
    for idx in range(seq_length - 1, -1, -1):
        best_tags = torch.gather(history_idx[idx], 1, best_tags)
        best_tags_arr[idx] = best_tags.data.view(batch_size)
    return best_tags_arr


@DECODERS.register_module(module_name=Decoders.constrained_crf)
class CRFwithConstraints(Decoder):
    """Conditional random field.
    This module implements a conditional random field [LMP01]_. The forward computation
    of this class computes the log likelihood of the given sequence of tags and
    emission score tensor. This class also has `~CRF.decode` method which finds
    the best tag sequence given an emission score tensor using `Viterbi algorithm`_.
    Args:
        num_tags: Number of tags.
        batch_first: Whether the first dimension corresponds to the size of a minibatch.
    Attributes:
        start_transitions (`~torch.nn.Parameter`): Start transition score tensor of size
            ``(num_tags,)``.
        end_transitions (`~torch.nn.Parameter`): End transition score tensor of size
            ``(num_tags,)``.
        transitions (`~torch.nn.Parameter`): Transition score tensor of size
            ``(num_tags, num_tags)``.
    .. [LMP01] Lafferty, J., McCallum, A., Pereira, F. (2001).
       "Conditional random fields: Probabilistic models for segmenting and
       labeling sequence data". *Proc. 18th International Conf. on Machine
       Learning*. Morgan Kaufmann. pp. 282–289.
    .. _Viterbi algorithm: https://en.wikipedia.org/wiki/Viterbi_algorithm

    Modified CRF.
    The only difference with CRF are the parameters: num_tags -> label2id and add constarints
    Tempororaily class, which will be merged into CRF in the future.
    """

    def __init__(
        self, id2label: List[str], batch_first: bool = False, add_constraint: bool = True
    ) -> None:
        num_tags = len(id2label)
        if num_tags <= 0:
            raise ValueError(f'invalid number of tags: {num_tags}')
        super().__init__()
        self.num_tags = num_tags
        self.batch_first = batch_first
        self.start_transitions = nn.Parameter(torch.empty(num_tags))
        self.end_transitions = nn.Parameter(torch.empty(num_tags))
        self.transitions = nn.Parameter(torch.empty(num_tags, num_tags))

        self.id2label = id2label
        self.label2id = {label: id for id, label in enumerate(id2label)}

        self.add_constraint = add_constraint

        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Initialize the transition parameters.
        The parameters will be initialized randomly from a uniform distribution
        between -0.1 and 0.1.
        """
        nn.init.uniform_(self.start_transitions, -0.1, 0.1)
        nn.init.uniform_(self.end_transitions, -0.1, 0.1)
        nn.init.uniform_(self.transitions, -0.1, 0.1)
        is_bio = sum([(i[:2] in ['E-', 'S-']) for i in self.id2label]) == 0
        if self.add_constraint:
            if is_bio:
                self.add_constraint_for_iob()
            else:
                self.add_constraint_for_iobes()

    def add_constraint_for_iobes(self):  # noqa
        print('[Info] Adding IOBES constraints')
        # add constraint:
        for prev_label in self.id2label:
            for next_label in self.id2label:
                if prev_label == 'O' and (next_label.startswith('I') or next_label.startswith('E')):
                    self.transitions.data[
                        self.label2id[prev_label], self.label2id[next_label]
                    ] = -10000.0
                if prev_label.startswith('B') or prev_label.startswith('I'):
                    if (
                        next_label.startswith('O')
                        or next_label.startswith('B')
                        or next_label.startswith('S')
                    ):
                        self.transitions.data[
                            self.label2id[prev_label], self.label2id[next_label]
                        ] = -10000.0
                    elif prev_label[2:] != next_label[2:]:
                        self.transitions.data[
                            self.label2id[prev_label], self.label2id[next_label]
                        ] = -10000.0
                if prev_label.startswith('S') or prev_label.startswith('E'):
                    if next_label.startswith('I') or next_label.startswith('E'):
                        self.transitions.data[
                            self.label2id[prev_label], self.label2id[next_label]
                        ] = -10000.0

        # constraint for start and end
        for label in self.id2label:
            if label.startswith('I') or label.startswith('E'):
                self.start_transitions.data[self.label2id[label]] = -10000.0
            if label.startswith('I') or label.startswith('B'):
                self.end_transitions.data[self.label2id[label]] = -10000.0

    def add_constraint_for_iob(self):  # noqa
        print('[Info] Adding IOB constraints')
        # add constraint:
        for prev_label in self.id2label:
            for next_label in self.id2label:
                if prev_label == 'O' and (next_label.startswith('I-')):
                    self.transitions.data[
                        self.label2id[prev_label], self.label2id[next_label]
                    ] = -10000.0

        # constraint for start and end
        for label in self.id2label:
            if label.startswith('I-'):
                self.start_transitions.data[self.label2id[label]] = -10000.0

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(num_tags={self.num_tags})'

    def forward(
        self,
        emissions: torch.Tensor,
        tags: torch.LongTensor,
        mask: Optional[torch.ByteTensor] = None,
        reduction: str = 'mean',
    ) -> torch.Tensor:
        """Compute the conditional log likelihood of a sequence of tags given emission scores.
        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length, num_tags)`` otherwise.
            tags (`~torch.LongTensor`): Sequence of tags tensor of size
                ``(seq_length, batch_size)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length)`` otherwise.
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
                if ``batch_first`` is ``False``, ``(batch_size, seq_length)`` otherwise.
            reduction: Specifies  the reduction to apply to the output:
                ``none|sum|mean|token_mean``. ``none``: no reduction will be applied.
                ``sum``: the output will be summed over batches. ``mean``: the output will be
                averaged over batches. ``token_mean``: the output will be averaged over tokens.
        Returns:
            `~torch.Tensor`: The log likelihood. This will have size ``(batch_size,)`` if
            reduction is ``none``, ``()`` otherwise.
        """
        if reduction not in ('none', 'sum', 'mean', 'token_mean'):
            raise ValueError(f'invalid reduction: {reduction}')
        if mask is None:
            mask = torch.ones_like(tags, dtype=torch.uint8, device=tags.device)
        if mask.dtype != torch.uint8:
            mask = mask.byte()
        self._validate(emissions, tags=tags, mask=mask)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            tags = tags.transpose(0, 1)
            mask = mask.transpose(0, 1)

        # shape: (batch_size,)
        numerator = self._compute_score(emissions, tags, mask)
        # shape: (batch_size,)
        denominator = self._compute_normalizer(emissions, mask)
        # shape: (batch_size,)
        llh = numerator - denominator

        if reduction == 'none':
            return llh
        if reduction == 'sum':
            return llh.sum()
        if reduction == 'mean':
            return llh.mean()
        return llh.sum() / mask.float().sum()

    def decode(
        self,
        emissions: torch.Tensor,
        mask: Optional[torch.ByteTensor] = None,
        nbest: Optional[int] = None,
        pad_tag: Optional[int] = None,
    ) -> List[List[List[int]]]:
        """Find the most likely tag sequence using Viterbi algorithm.
        Args:
            emissions (`~torch.Tensor`): Emission score tensor of size
                ``(seq_length, batch_size, num_tags)`` if ``batch_first`` is ``False``,
                ``(batch_size, seq_length, num_tags)`` otherwise.
            mask (`~torch.ByteTensor`): Mask tensor of size ``(seq_length, batch_size)``
                if ``batch_first`` is ``False``, ``(batch_size, seq_length)`` otherwise.
            nbest (`int`): Number of most probable paths for each sequence
            pad_tag (`int`): Tag at padded positions. Often input varies in length and
                the length will be padded to the maximum length in the batch. Tags at
                the padded positions will be assigned with a padding tag, i.e. `pad_tag`
        Returns:
            A PyTorch tensor of the best tag sequence for each batch of shape
            (nbest, batch_size, seq_length)
        """
        if nbest is None:
            nbest = 1
        if mask is None:
            mask = torch.ones(emissions.shape[:2], dtype=torch.uint8, device=emissions.device)
        if mask.dtype != torch.uint8:
            mask = mask.byte()
        self._validate(emissions, mask=mask)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            mask = mask.transpose(0, 1)

        if nbest == 1:
            return self._viterbi_decode(emissions, mask, pad_tag).unsqueeze(0)
        return self._viterbi_decode_nbest(emissions, mask, nbest, pad_tag)

    def _validate(
        self,
        emissions: torch.Tensor,
        tags: Optional[torch.LongTensor] = None,
        mask: Optional[torch.ByteTensor] = None,
    ) -> None:
        if emissions.dim() != 3:
            raise ValueError(f'emissions must have dimension of 3, got {emissions.dim()}')
        if emissions.size(2) != self.num_tags:
            raise ValueError(
                f'expected last dimension of emissions is {self.num_tags}, '
                f'got {emissions.size(2)}'
            )

        if tags is not None:
            if emissions.shape[:2] != tags.shape:
                raise ValueError(
                    'the first two dimensions of emissions and tags must match, '
                    f'got {tuple(emissions.shape[:2])} and {tuple(tags.shape)}'
                )

        if mask is not None:
            if emissions.shape[:2] != mask.shape:
                breakpoint()
                raise ValueError(
                    'the first two dimensions of emissions and mask must match, '
                    f'got {tuple(emissions.shape[:2])} and {tuple(mask.shape)}'
                )
            no_empty_seq = not self.batch_first and mask[0].all()
            no_empty_seq_bf = self.batch_first and mask[:, 0].all()
            if not no_empty_seq and not no_empty_seq_bf:
                raise ValueError('mask of the first timestep must all be on')

    def _compute_score(
        self, emissions: torch.Tensor, tags: torch.LongTensor, mask: torch.ByteTensor
    ) -> torch.Tensor:
        # emissions: (seq_length, batch_size, num_tags)
        # tags: (seq_length, batch_size)
        # mask: (seq_length, batch_size)
        seq_length, batch_size = tags.shape
        mask = mask.float()

        # Start transition score and first emission
        # shape: (batch_size,)
        score = self.start_transitions[tags[0]]
        score += emissions[0, torch.arange(batch_size), tags[0]]

        for i in range(1, seq_length):
            # Transition score to next tag, only added if next timestep is valid (mask == 1)
            # shape: (batch_size,)
            score += self.transitions[tags[i - 1], tags[i]] * mask[i]

            # Emission score for next tag, only added if next timestep is valid (mask == 1)
            # shape: (batch_size,)
            score += emissions[i, torch.arange(batch_size), tags[i]] * mask[i]

        # End transition score
        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1
        # shape: (batch_size,)
        last_tags = tags[seq_ends, torch.arange(batch_size)]
        # shape: (batch_size,)
        score += self.end_transitions[last_tags]

        return score

    def _compute_normalizer(self, emissions: torch.Tensor, mask: torch.ByteTensor) -> torch.Tensor:
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        seq_length = emissions.size(0)

        # Start transition score and first emission; score has size of
        # (batch_size, num_tags) where for each batch, the j-th column stores
        # the score that the first timestep has tag j
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]

        for i in range(1, seq_length):
            # Broadcast score for every possible next tag
            # shape: (batch_size, num_tags, 1)
            broadcast_score = score.unsqueeze(2)

            # Broadcast emission score for every possible current tag
            # shape: (batch_size, 1, num_tags)
            broadcast_emissions = emissions[i].unsqueeze(1)

            # Compute the score tensor of size (batch_size, num_tags, num_tags) where
            # for each sample, entry at row i and column j stores the sum of scores of all
            # possible tag sequences so far that end with transitioning from tag i to tag j
            # and emitting
            # shape: (batch_size, num_tags, num_tags)
            next_score = broadcast_score + self.transitions + broadcast_emissions

            # Sum over all possible current tags, but we're in score space, so a sum
            # becomes a log-sum-exp: for each sample, entry i stores the sum of scores of
            # all possible tag sequences so far, that end in tag i
            # shape: (batch_size, num_tags)
            next_score = torch.logsumexp(next_score, dim=1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # shape: (batch_size, num_tags)
            score = torch.where(mask[i].unsqueeze(1), next_score, score)

        # End transition score
        # shape: (batch_size, num_tags)
        score += self.end_transitions

        # Sum (log-sum-exp) over all possible tags
        # shape: (batch_size,)
        return torch.logsumexp(score, dim=1)

    def _viterbi_decode(
        self, emissions: torch.FloatTensor, mask: torch.ByteTensor, pad_tag: Optional[int] = None
    ) -> List[List[int]]:
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        # return: (batch_size, seq_length)
        if pad_tag is None:
            pad_tag = 0

        device = emissions.device
        seq_length, batch_size = mask.shape

        # Start transition and first emission
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]
        history_idx = torch.zeros(
            (seq_length, batch_size, self.num_tags), dtype=torch.long, device=device
        )
        oor_idx = torch.zeros((batch_size, self.num_tags), dtype=torch.long, device=device)
        oor_tag = torch.full((seq_length, batch_size), pad_tag, dtype=torch.long, device=device)

        # - score is a tensor of size (batch_size, num_tags) where for every batch,
        #   value at column j stores the score of the best tag sequence so far that ends
        #   with tag j
        # - history_idx saves where the best tags candidate transitioned from; this is used
        #   when we trace back the best tag sequence
        # - oor_idx saves the best tags candidate transitioned from at the positions
        #   where mask is 0, i.e. out of range (oor)

        # Viterbi algorithm recursive case: we compute the score of the best tag sequence
        # for every possible next tag
        for i in range(1, seq_length):
            # Broadcast viterbi score for every possible next tag
            # shape: (batch_size, num_tags, 1)
            broadcast_score = score.unsqueeze(2)

            # Broadcast emission score for every possible current tag
            # shape: (batch_size, 1, num_tags)
            broadcast_emission = emissions[i].unsqueeze(1)

            # Compute the score tensor of size (batch_size, num_tags, num_tags) where
            # for each sample, entry at row i and column j stores the score of the best
            # tag sequence so far that ends with transitioning from tag i to tag j and emitting
            # shape: (batch_size, num_tags, num_tags)
            next_score = broadcast_score + self.transitions + broadcast_emission

            # Find the maximum score over all possible current tag
            # shape: (batch_size, num_tags)
            next_score, indices = next_score.max(dim=1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # and save the index that produces the next score
            # shape: (batch_size, num_tags)
            score = torch.where(mask[i].unsqueeze(-1), next_score, score)
            indices = torch.where(mask[i].unsqueeze(-1), indices, oor_idx)
            history_idx[i - 1] = indices

        # End transition score
        # shape: (batch_size, num_tags)
        end_score = score + self.end_transitions
        _, end_tag = end_score.max(dim=1)

        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1

        # insert the best tag at each sequence end (last position with mask == 1)
        history_idx = history_idx.transpose(1, 0).contiguous()
        history_idx.scatter_(
            1,
            seq_ends.view(-1, 1, 1).expand(-1, 1, self.num_tags),
            end_tag.view(-1, 1, 1).expand(-1, 1, self.num_tags),
        )
        history_idx = history_idx.transpose(1, 0).contiguous()

        # The most probable path for each sequence
        best_tags_arr = torch.zeros((seq_length, batch_size), dtype=torch.long, device=device)
        best_tags = torch.zeros(batch_size, 1, dtype=torch.long, device=device)
        for idx in range(seq_length - 1, -1, -1):
            best_tags = torch.gather(history_idx[idx], 1, best_tags)
            best_tags_arr[idx] = best_tags.data.view(batch_size)

        return torch.where(mask, best_tags_arr, oor_tag).transpose(0, 1)

    def _viterbi_decode_nbest(
        self,
        emissions: torch.FloatTensor,
        mask: torch.ByteTensor,
        nbest: int,
        pad_tag: Optional[int] = None,
    ) -> List[List[List[int]]]:
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        # return: (nbest, batch_size, seq_length)
        if pad_tag is None:
            pad_tag = 0

        device = emissions.device
        seq_length, batch_size = mask.shape

        # Start transition and first emission
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]
        history_idx = torch.zeros(
            (seq_length, batch_size, self.num_tags, nbest), dtype=torch.long, device=device
        )
        oor_idx = torch.zeros((batch_size, self.num_tags, nbest), dtype=torch.long, device=device)
        oor_tag = torch.full(
            (seq_length, batch_size, nbest), pad_tag, dtype=torch.long, device=device
        )

        # + score is a tensor of size (batch_size, num_tags) where for every batch,
        #   value at column j stores the score of the best tag sequence so far that ends
        #   with tag j
        # + history_idx saves where the best tags candidate transitioned from; this is used
        #   when we trace back the best tag sequence
        # - oor_idx saves the best tags candidate transitioned from at the positions
        #   where mask is 0, i.e. out of range (oor)

        # Viterbi algorithm recursive case: we compute the score of the best tag sequence
        # for every possible next tag
        for i in range(1, seq_length):
            if i == 1:
                broadcast_score = score.unsqueeze(-1)
                broadcast_emission = emissions[i].unsqueeze(1)
                # shape: (batch_size, num_tags, num_tags)
                next_score = broadcast_score + self.transitions + broadcast_emission
            else:
                broadcast_score = score.unsqueeze(-1)
                broadcast_emission = emissions[i].unsqueeze(1).unsqueeze(2)
                # shape: (batch_size, num_tags, nbest, num_tags)
                next_score = broadcast_score + self.transitions.unsqueeze(1) + broadcast_emission

            # Find the top `nbest` maximum score over all possible current tag
            # shape: (batch_size, nbest, num_tags)
            next_score, indices = next_score.view(batch_size, -1, self.num_tags).topk(nbest, dim=1)

            if i == 1:
                score = score.unsqueeze(-1).expand(-1, -1, nbest)
                indices = indices * nbest

            # convert to shape: (batch_size, num_tags, nbest)
            next_score = next_score.transpose(2, 1)
            indices = indices.transpose(2, 1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # and save the index that produces the next score
            # shape: (batch_size, num_tags, nbest)
            score = torch.where(mask[i].unsqueeze(-1).unsqueeze(-1), next_score, score)
            indices = torch.where(mask[i].unsqueeze(-1).unsqueeze(-1), indices, oor_idx)
            history_idx[i - 1] = indices

        # End transition score shape: (batch_size, num_tags, nbest)
        end_score = score + self.end_transitions.unsqueeze(-1)
        _, end_tag = end_score.view(batch_size, -1).topk(nbest, dim=1)

        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1

        # insert the best tag at each sequence end (last position with mask == 1)
        history_idx = history_idx.transpose(1, 0).contiguous()
        history_idx.scatter_(
            1,
            seq_ends.view(-1, 1, 1, 1).expand(-1, 1, self.num_tags, nbest),
            end_tag.view(-1, 1, 1, nbest).expand(-1, 1, self.num_tags, nbest),
        )
        history_idx = history_idx.transpose(1, 0).contiguous()

        # The most probable path for each sequence
        best_tags_arr = torch.zeros(
            (seq_length, batch_size, nbest), dtype=torch.long, device=device
        )
        best_tags = (
            torch.arange(nbest, dtype=torch.long, device=device).view(1, -1).expand(batch_size, -1)
        )
        for idx in range(seq_length - 1, -1, -1):
            best_tags = torch.gather(history_idx[idx].view(batch_size, -1), 1, best_tags)
            best_tags_arr[idx] = best_tags.data.view(batch_size, -1) // nbest

        return torch.where(mask.unsqueeze(-1), best_tags_arr, oor_tag).permute(2, 1, 0)



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, List, Union

from modelscope.preprocessors.builder import PREPROCESSORS
from modelscope.utils.constant import Fields

from adaseq.metainfo import Preprocessors

from .pretraining_preprocessor import PretrainingPreprocessor


@PREPROCESSORS.register_module(Fields.nlp, module_name=Preprocessors.twostage_preprocessor)
class TwoStagePreprocessor(PretrainingPreprocessor):
    """Preprocessor for twostage-ner.
    span targets are processed into mention_boundary, type_ids, ident_ids.
    examples:
        span: {'start':1, 'end':2, 'type': ['PER']}
        processed: {'mention_boundary': [[1], [2]], 'type_ids':[1], 'ident_ids': ['S-SPAN']]}
    """

    def _generate_mentions_labels(self, data: Union[str, List, Dict], output: Dict[str, Any]):
        mention_type_ids = []
        boundary_starts = []
        boundary_ends = []
        mention_mask = []
        for span in data['spans']:
            boundary_starts.append(span['start'])
            boundary_ends.append(span['end'] - 1)
            mention_mask.append(1)
            mention_type_ids.append(self.typing_label_to_id.get(span['type'], -1))

        output['mention_boundary'] = [boundary_starts, boundary_ends]
        output['mention_mask'] = mention_mask
        output['type_ids'] = mention_type_ids



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict

from modelscope.preprocessors.builder import PREPROCESSORS
from modelscope.utils.constant import Fields

from adaseq.metainfo import Preprocessors

from .nlp_preprocessor import NLPPreprocessor


@PREPROCESSORS.register_module(
    Fields.nlp, module_name=Preprocessors.multilabel_span_typing_preprocessor
)
class MultiLabelSpanTypingPreprocessor(NLPPreprocessor):
    """Preprocessor for multilabel (aka multi-type) span typing task.
    span targets are processed into mention_boundary, type_ids.
    examples:
        span: {'start':1, 'end':2, 'type': ['PER']}
        processed: {'mention_boundary': [[1], [2]], 'type_ids':[1]}
    """

    def __init__(self, model_dir: str, **kwargs):
        super().__init__(model_dir, return_offsets=True, **kwargs)

    def __call__(self, data: Dict) -> Dict[str, Any]:
        """prepare inputs for Entity Typing model"""
        output = super().__call__(data)

        mention_type_ids = []
        boundary_starts = []
        boundary_ends = []
        mention_mask = []
        for span in data['spans']:
            boundary_starts.append(span['start'])
            boundary_ends.append(span['end'] - 1)
            mention_mask.append(1)
            type_ids = [self.label_to_id.get(x, -1) for x in span['type']]
            padded_type_ids = [0] * len(self.label_to_id)
            for t in type_ids:
                if t != -1:
                    padded_type_ids[t] = 1
            mention_type_ids.append(padded_type_ids)

        output['mention_boundary'] = [boundary_starts, boundary_ends]
        output['mention_mask'] = mention_mask  # mask, 为了避开nlp_preprocessor对他做padding.
        output['type_ids'] = mention_type_ids
        return output


@PREPROCESSORS.register_module(module_name=Preprocessors.multilabel_concat_typing_preprocessor)
class MultiLabelConcatTypingPreprocessor(NLPPreprocessor):
    """Preprocessor for multilabel (aka multi-type) span concat typing task."""

    def __init__(self, model_dir: str, **kwargs):
        super().__init__(model_dir, return_offsets=True, **kwargs)

    # label_boundary:
    #    in: [{'start': s, 'end': e, 'types': l}]
    #    out: [[s,e]]
    # label_ids:
    #    in: [{'start': s, 'end': e, 'types': l}]
    #    out: [[0,1,0]] # [[*] * num_classes(one-hot type vector)]
    def __call__(self, data: Dict) -> Dict[str, Any]:  # noqa: D102
        spans = data.get('spans', [])
        assert len(spans) == 1, 'ConcatTyping only supports single mention per data'
        span = spans[0]
        tokens = data['tokens']
        type_ids = [self.label_to_id.get(x, -1) for x in span['type']]
        padded_type_ids = [0] * len(self.label_to_id)  # multilabel type id onehot vector
        for t in type_ids:
            if t != -1:
                padded_type_ids[t] = 1

        mention = tokens[span['start'] : span['end']]  # since end is open
        sent = (
            [self.tokenizer.cls_token]
            + tokens
            + [self.tokenizer.sep_token]
            + mention
            + [self.tokenizer.sep_token]
        )
        input_ids = []
        for tok in sent:
            subtoken_ids = self.tokenizer.encode(tok, add_special_tokens=False)
            if len(subtoken_ids) == 0:
                subtoken_ids = [self.tokenizer.unk_token_id]
            input_ids.extend(subtoken_ids)

        if len(input_ids) > self.max_length:
            input_ids = [input_ids[0]] + input_ids[-(self.max_length - 1) :]
            # clip from the left except cls token

        attention_mask = [1] * len(input_ids)
        output = {
            'tokens': {
                'input_ids': input_ids,
                'attention_mask': attention_mask,
            },
            'type_ids': padded_type_ids,
            'meta': data,
        }
        return output



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Optional

import torch
import torch.nn as nn

from adaseq.metainfo import Encoders

from .base import ENCODERS, Encoder


@ENCODERS.register_module(module_name=Encoders.span_encoder)
class SpanEncoder(Encoder):
    """Turn token embedding sequenece to a single vector."""

    def __init__(
        self,
        input_dim: int,
        encode_span_method: str = 'concat',
        add_span_linear: bool = True,
        span_hidden_size: Optional[int] = None,
        use_biaffine: bool = False,
        **kwargs
    ):
        super().__init__()
        self.input_dim = input_dim
        self.add_span_linear = add_span_linear
        self.encode_span_method = encode_span_method
        if self.add_span_linear:
            self.span_hidden_size = input_dim
            if span_hidden_size is not None:
                self.span_hidden_size = span_hidden_size
            self.start_linear_mapping = nn.Linear(input_dim, self.span_hidden_size)
            self.end_linear_mapping = nn.Linear(input_dim, self.span_hidden_size)
        self.use_biaffine = use_biaffine
        if self.use_biaffine:
            if self.add_span_linear:
                self.output_dim = self.span_hidden_size
            else:
                self.output_dim = self.input_dim
        elif encode_span_method == 'concat':
            if self.add_span_linear:
                self.span_reprs_dim = 2 * self.span_hidden_size
                self.output_dim = self.span_reprs_dim
            else:
                self.span_reprs_dim = 2 * self.input_dim
                self.output_dim = self.span_reprs_dim

    def get_input_dim(self) -> int:  # noqa: D102
        return self.input_dim

    def get_output_dim(self) -> int:  # noqa: D102
        return self.output_dim

    def forward(self, token_embed, span_boundary):  # noqa
        # B x N x K -> select accroding to span_boundary -> B x M x K
        batch_size = span_boundary.shape[0]
        token_embed = token_embed.reshape(batch_size, -1, self.input_dim)
        span_start_idx = span_boundary[:, 0, :]
        span_end_idx = span_boundary[:, 1, :]
        span_start_embed = token_embed[torch.arange(batch_size)[:, None], span_start_idx]
        span_end_embed = token_embed[torch.arange(batch_size)[:, None], span_end_idx]
        if self.add_span_linear:
            span_start_reprs = self.start_linear_mapping(span_start_embed)
            span_end_reprs = self.end_linear_mapping(span_end_embed)
        else:
            span_start_reprs = span_start_embed
            span_end_reprs = span_end_embed
        if self.use_biaffine:
            return span_start_reprs.reshape(-1, self.span_hidden_size), span_end_reprs.reshape(
                -1, self.span_hidden_size
            )
        if self.encode_span_method == 'concat':
            span_reprs = torch.cat((span_start_reprs, span_end_reprs), -1)
        else:
            raise NotImplementedError
        span_reprs = span_reprs.reshape(-1, self.span_reprs_dim)
        return span_reprs



# Copyright (c) Alibaba, Inc. and its affiliates.
from .pretraining_metric import PretrainingMetric
from .relation_extraction_metric import RelationExtractionMetric
from .sequence_labeling_metric import SequenceLabelingMetric
from .span_extraction_metric import SpanExtractionMetric
from .typing_metric import TypingMetric



# Copyright (c) Alibaba, Inc. and its affiliates.

from typing import Any, Dict, Optional, Union

import torch
from modelscope.models import Model
from modelscope.pipelines.base import Pipeline as MsPipeline
from modelscope.preprocessors import Preprocessor


class Pipeline(MsPipeline):
    """use `model` and `preprocessor` to create a pipeline for prediction

    Args:
        model (str or Model): A model instance or a model local dir or a model id in the model hub.
        preprocessor (Preprocessor): a preprocessor instance, must not be None.
        kwargs (dict, `optional`):
            Extra kwargs passed into the preprocessor's constructor.
    """

    def __init__(
        self,
        model: Union[Model, str],
        preprocessor: Optional[Preprocessor] = None,
        config_file: str = None,
        device: str = 'gpu',
        auto_collate: bool = True,
        **kwargs
    ):
        super().__init__(
            model=model,
            preprocessor=preprocessor,
            config_file=config_file,
            device=device,
            auto_collate=auto_collate,
        )

        if preprocessor is None:
            self.preprocessor = Preprocessor.from_pretrained(self.model.model_dir, **kwargs)
        self.model.eval()

        assert hasattr(self.preprocessor, 'id_to_label')
        self.id2label = self.preprocessor.id_to_label

    def forward(self, inputs: Dict[str, Any], **forward_params) -> Dict[str, Any]:  # noqa: D102
        with torch.no_grad():
            return {**self.model(**inputs, **forward_params), **inputs}

    def postprocess(  # noqa: D102
        self, inputs: Dict[str, Any], **postprocess_params
    ) -> Dict[str, Any]:
        raise NotImplementedError



# Copyright (c) Alibaba, Inc. and its affiliates.
import torch
import torch.nn as nn


class WordDropout(nn.Module):
    """Word-level Dropout module

    During training, randomly zeroes some of the elements of the input tensor at word level
    with probability `dropout_rate` using samples from a Bernoulli distribution.

    Args:
        dropout_rate (float): dropout rate for each word
    """

    def __init__(self, dropout_rate: float = 0.1):
        super(WordDropout, self).__init__()
        assert 0.0 <= dropout_rate < 1.0, '0.0 <= dropout rate < 1.0 must be satisfied!'
        self.dropout_rate = dropout_rate

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Dropout the input tensor at word level

        Args:
            inputs (torch.Tensor): input tensor

        Returns:
            outputs (torch.Tensor): output tensor of the same shape as input
        """
        if not self.training or not self.dropout_rate:
            return inputs

        mask = inputs.new_empty(*inputs.shape[:2], 1, requires_grad=False).bernoulli_(
            1.0 - self.dropout_rate
        )
        mask = mask.expand_as(inputs)
        return inputs * mask



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) AI2 AllenNLP. Licensed under the Apache License, Version 2.0.
from typing import Any, Dict, Optional, Tuple

import torch
import transformers
from modelscope.models import Model as MsModel
from transformers import AutoConfig, AutoModel, XLNetConfig

from adaseq.metainfo import Embedders
from adaseq.modules import util
from adaseq.modules.scalar_mix import ScalarMix

from .base import EMBEDDERS, Embedder


# Code partially borrowed from https://github.com/allenai/allennlp/blob/HEAD/
# allennlp/modules/token_embedders/pretrained_transformer_embedder.py
@EMBEDDERS.register_module(module_name=Embedders.transformer_embedder)
class TransformerEmbedder(Embedder):
    """
    Uses a pretrained model from `transformers` as a `Embedder`.

    # Parameters

    model_name_or_path : `str`
        The name of the `transformers` model to use. Should be the same as the corresponding
        `PretrainedTransformerIndexer`.
    drop_special_tokens:  `bool` (default = `True`)
        if `True`, drop the hidden states of special tokens (currently [CLS], [SEP]).
    sub_module: `str`, optional (default = `None`)
        The name of a submodule of the transformer to be used as the embedder. Some transformers naturally act
        as embedders such as BERT. However, other models consist of encoder and decoder, in which case we just
        want to use the encoder.
    train_parameters: `bool`, optional (default = `True`)
        If this is `True`, the transformer weights get updated during training. If this is `False`, the
        transformer weights are not updated during training.
    eval_mode: `bool`, optional (default = `False`)
        If this is `True`, the model is always set to evaluation mode (e.g., the dropout is disabled and the
        batch normalization layer statistics are not updated). If this is `False`, such dropout and batch
        normalization layers are only set to evaluation mode when when the model is evaluating on development
        or test data.
    load_weights: `bool`, optional (default = `True`)
        Whether to load the pretrained weights. If you're loading your model/predictor from an AllenNLP archive
        it usually makes sense to set this to `False` (via the `overrides` parameter)
        to avoid unnecessarily caching and loading the original pretrained weights,
        since the archive will already contain all of the weights needed.
    scalar_mix: `Dict[str, Any]`, optional (default = `None`)
        When `None` (the default), only the final layer of the pretrained transformer is taken
        for the embeddings.
        If pass a kwargs dict, a scalar mix of all of the layers is used.
    gradient_checkpointing: `bool`, optional (default = `None`)
        Enable or disable gradient checkpointing.
    transformer_kwargs: `Dict[str, Any]`, optional (default = `None`)
        Dictionary with additional arguments for `get_transformer`.
    sub_token_mode: `str`, (default= `first`)
        If `sub_token_mode` is set to `first`, return first sub-token representation as word-level representation.
        If `sub_token_mode` is set to `last`, return last sub-token representation as word-level representation.
        If `sub_token_mode` is set to `avg`, return average of all the sub-tokens
        representation as word-level representation.
        If `sub_token_mode` is not specified it defaults to `avg`
        If invalid `sub_token_mode` is provided, throw `ConfigurationError`
    """

    def __init__(
        self,
        model_name_or_path: str,
        *,
        drop_special_tokens: bool = True,
        sub_module: Optional[str] = None,
        train_parameters: bool = True,
        eval_mode: bool = False,
        load_weights: bool = True,
        scalar_mix: Optional[Dict[str, Any]] = None,
        gradient_checkpointing: Optional[bool] = None,
        transformer_kwargs: Optional[Dict[str, Any]] = None,
        sub_token_mode: str = 'first',
    ) -> None:
        super().__init__()
        self.drop_special_tokens = drop_special_tokens
        self.sub_token_mode = sub_token_mode

        self.transformer_model, self.from_hf = get_transformer(
            model_name_or_path,
            load_weights=load_weights,
            **(transformer_kwargs or {}),
        )

        if self.from_hf:
            if gradient_checkpointing is not None:
                self.transformer_model.config.update(
                    {'gradient_checkpointing': gradient_checkpointing}
                )

        self.config = self.transformer_model.config

        if sub_module:
            assert hasattr(self.transformer_model, sub_module)
            self.transformer_model = getattr(self.transformer_model, sub_module)

        # I'm not sure if this works for all models; open an issue on github if you find a case
        # where it doesn't work.
        self.output_dim = self.config.hidden_size

        self._scalar_mix: Optional[ScalarMix] = None
        if scalar_mix:
            self._scalar_mix = ScalarMix(self.config.num_hidden_layers, **scalar_mix)
            self.config.output_hidden_states = True

        self.train_parameters = train_parameters
        if not train_parameters:
            for param in self.transformer_model.parameters():
                param.requires_grad = False

        self.eval_mode = eval_mode
        if eval_mode:
            self.transformer_model.eval()

        if isinstance(self.config, XLNetConfig):
            self._number_of_token_type_embeddings = 3  # XLNet has 3 type ids
        elif hasattr(self.config, 'type_vocab_size'):
            self._number_of_token_type_embeddings = self.config.type_vocab_size
        else:
            self._number_of_token_type_embeddings = 0

    def train(self, mode: bool = True):  # noqa: D102
        self.training = mode
        for name, module in self.named_children():
            if self.eval_mode and name == 'transformer_model':
                module.eval()
            else:
                module.train(mode)
        return self

    def get_output_dim(self):  # noqa: D102
        return self.output_dim

    def forward(
        self,
        input_ids: torch.LongTensor,
        attention_mask: torch.BoolTensor,
        token_type_ids: Optional[torch.LongTensor] = None,
        has_special_tokens: Optional[torch.BoolTensor] = None,
        offsets: Optional[torch.LongTensor] = None,
        mask: Optional[torch.BoolTensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """
        input_ids: `torch.LongTensor`
            Shape: [batch_size, num_orig_tokens or num_wordpieces].
        attention_mask: `torch.BoolTensor`
            Shape: [batch_size, num_orig_tokens or num_wordpieces].
        token_type_ids: `Optional[torch.LongTensor]`
            Shape: [batch_size, num_orig_tokens or num_wordpieces].
        offsets: `torch.LongTensor`
            Shape: [batch_size, num_orig_tokens, 2].
            Maps indices for the original tokens, i.e. those given as input to the indexer,
            to a span in input_ids. `input_ids[i][offsets[i][j][0]:offsets[i][j][1] + 1]`
            corresponds to the original j-th token from the i-th batch.
        mask: `torch.BoolTensor`
            Shape: [batch_size, num_orig_tokens].

        # Returns

        `torch.Tensor`
            Shape: `[batch_size, num_wordpieces or num_orig_tokens, embedding_size]`.
        """
        # first encode sub-token level representations
        encoded = self.encode(input_ids, attention_mask, token_type_ids)  # type: ignore
        if offsets is not None:
            # then reconstruct token-level ones by offsets
            encoded = self.reconstruct(encoded, offsets)

        if has_special_tokens is not None:
            if self.drop_special_tokens and has_special_tokens.bool()[0]:
                encoded = encoded[:, 1:-1]  # So far, we only consider [CLS] and [SEP]
        return encoded

    def encode(
        self,
        input_ids: torch.LongTensor,
        attention_mask: torch.BoolTensor,
        token_type_ids: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:  # type: ignore
        """
        # Parameters

        input_ids: `torch.LongTensor`
            Shape: `[batch_size, num_wordpieces]`.
        attention_mask: `torch.BoolTensor`
            Shape: [batch_size, num_wordpieces].
        token_type_ids: `Optional[torch.LongTensor]`
            Shape: `[batch_size, num_wordpieces]`.

        # Returns

        `torch.Tensor`
            Shape: `[batch_size, num_wordpieces, embedding_size]`.

        """
        # Some of the huggingface transformers don't support type ids at all and crash when you supply
        # them. For others, you can supply a tensor of zeros, and if you don't, they act as if you did.
        # There is no practical difference to the caller, so here we pretend that one case is the same
        # as another case.
        if token_type_ids is not None:
            max_type_id = token_type_ids.max()
            if max_type_id == 0:
                token_type_ids = None
            else:
                if max_type_id >= self._number_of_token_type_embeddings:
                    raise ValueError('Found type ids too large for the chosen transformer model.')
                assert input_ids.shape == token_type_ids.shape

        assert attention_mask is not None
        # Shape: [batch_size, num_wordpieces, embedding_size],

        # We call this with kwargs because some of the huggingface models don't have the
        # token_type_ids parameter and fail even when it's given as None.
        # Also, as of transformers v2.5.1, they are taking FloatTensor masks.
        parameters = {'input_ids': input_ids, 'attention_mask': attention_mask.float()}
        if token_type_ids is not None:
            parameters['token_type_ids'] = token_type_ids

        transformer_output = self.transformer_model(**parameters)
        if self._scalar_mix is not None:
            # The hidden states will also include the embedding layer, which we don't
            # include in the scalar mix. Hence the `[1:]` slicing.
            hidden_states = transformer_output.hidden_states[1:]
            embeddings = self._scalar_mix(hidden_states)
        else:
            embeddings = transformer_output.last_hidden_state

        return embeddings

    def reconstruct(self, embeddings: torch.Tensor, offsets: torch.LongTensor) -> torch.Tensor:
        """
        # Parameters

        input_ids: `torch.LongTensor`
            Shape: [batch_size, num_wordpieces].
        offsets: `torch.LongTensor`
            Shape: [batch_size, num_orig_tokens, 2].
            Maps indices for the original tokens, i.e. those given as input to the indexer,
            to a span in input_ids. `input_ids[i][offsets[i][j][0]:offsets[i][j][1] + 1]`
            corresponds to the original j-th token from the i-th batch.

        # Returns

        `torch.Tensor`
            Shape: [batch_size, num_orig_tokens, embedding_size].
        """

        # If "sub_token_mode" is set to "first", return the first sub-token embedding
        if self.sub_token_mode == 'first':
            # Select first sub-token embeddings from span embeddings
            # Shape: (batch_size, num_orig_tokens, embedding_size)
            orig_embeddings = util.batched_index_select(embeddings, offsets[..., 0])

        # If "sub_token_mode" is set to "last", return the last sub-token embedding
        elif self.sub_token_mode == 'last':
            # Select last sub-token embeddings from span embeddings
            # Shape: (batch_size, num_orig_tokens, embedding_size)
            orig_embeddings = util.batched_index_select(embeddings, offsets[..., 1])

        # If "sub_token_mode" is set to "avg", return the average of embeddings of all sub-tokens of a word
        elif self.sub_token_mode == 'avg':
            # span_embeddings: (batch_size, num_orig_tokens, max_span_length, embedding_size)
            # span_mask: (batch_size, num_orig_tokens, max_span_length)
            span_embeddings, span_mask = util.batched_span_select(embeddings.contiguous(), offsets)

            span_mask = span_mask.unsqueeze(-1)

            # Shape: (batch_size, num_orig_tokens, max_span_length, embedding_size)
            span_embeddings *= span_mask  # zero out paddings

            # Sum over embeddings of all sub-tokens of a word
            # Shape: (batch_size, num_orig_tokens, embedding_size)
            span_embeddings_sum = span_embeddings.sum(2)

            # Shape (batch_size, num_orig_tokens)
            span_embeddings_len = span_mask.sum(2)

            # Find the average of sub-tokens embeddings by dividing `span_embedding_sum` by `span_embedding_len`
            # Shape: (batch_size, num_orig_tokens, embedding_size)
            orig_embeddings = span_embeddings_sum / torch.clamp_min(span_embeddings_len, 1)

            # All the places where the span length is zero, write in zeros.
            orig_embeddings[(span_embeddings_len == 0).expand(orig_embeddings.shape)] = 0

        # If invalid "sub_token_mode" is provided, throw error
        else:
            raise ValueError(f"Do not recognise 'sub_token_mode' {self.sub_token_mode}")

        return orig_embeddings


def get_transformer(
    model_name_or_path: str,
    load_weights: bool = True,
    source: Optional[str] = None,
    **kwargs,
) -> Tuple[transformers.PreTrainedModel, bool]:
    """
    Returns a transformer model and a flag of whether comes from huggingface.

    # Parameters

    model_name_or_path : `str`
        The name of the transformer, for example `"bert-base-cased"`
    load_weights : `bool`, optional (default = `True`)
        If set to `False`, no weights will be loaded. This is helpful when you only
        want to initialize the architecture, like when you've already fine-tuned a model
        and are going to load the weights from a state dict elsewhere.
        Only effective when loading model from huggingface.
    source : `str`, optional (default = `None`),
        if `source == 'huggingface'`, only try to load from huggingface.
        `source == 'modelscope'` is similar.
        By default, we will try huggingface first, and then modelscope, if both failed,
        TODO
    kwargs: `Dict[str, Any]`, optional (default = `None`)
        Dictionary with additional arguments for `XxxxModel.from_pretrained`.

    # Returns
        Tuple[transformers.PreTrainedModel, bool], the transformer and a flag, which
        is `True` when successfully loaded from huggingface (`False` for modelscope)
    """
    from requests.exceptions import HTTPError

    from adaseq.utils.checks import ConfigurationError

    if isinstance(source, str):
        if source.lower() == 'hugginface':
            return get_hf_transformer(model_name_or_path, load_weights, **kwargs), True

        elif source.lower() == 'modelscope':
            return get_ms_transformer(model_name_or_path, **kwargs), False

        else:
            raise ConfigurationError(f'Unsupported transformer source: {source}')

    elif source is None:
        hf_e, ms_e = None, None

        try:
            return get_hf_transformer(model_name_or_path, load_weights, **kwargs), True
        except OSError as e:
            hf_e = e

        try:
            return get_ms_transformer(model_name_or_path, **kwargs), False
        except HTTPError as e:
            ms_e = e

        message = 'Try loading from huggingface and modelscope failed \n\n'
        message += 'huggingface:\n' + str(hf_e)
        message += '\n\nmodelscope:\n' + str(ms_e)
        raise RuntimeError(message)

    else:
        raise ConfigurationError(f'Unsupported transformer source: {source}')


def get_hf_transformer(
    model_name_or_path: str,
    load_weights: bool = True,
    **kwargs,
) -> transformers.PreTrainedModel:
    """see `get_transformer`."""
    if load_weights:
        transformer = AutoModel.from_pretrained(model_name_or_path, **kwargs)
    else:
        transformer = AutoModel.from_config(
            AutoConfig.from_pretrained(model_name_or_path, **kwargs)
        )
    return transformer


def get_ms_transformer(
    model_name_or_path: str,
    **kwargs,
) -> transformers.PreTrainedModel:
    """see `get_transformer`."""
    from adaseq.utils.checks import ConfigurationError

    try:
        transformer = MsModel.from_pretrained(model_name_or_path, task='backbone', **kwargs)
    except KeyError:
        transformer = MsModel.from_pretrained(model_name_or_path, **kwargs)

        # only support transformer-crf model for now
        model_type = transformer.cfg.model.type
        if model_type == 'transformer-crf':
            transformer = transformer.model.encoder
        else:
            raise ConfigurationError(f'Unsupported non-backbone embedder: {model_name_or_path}')

    return transformer



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Dict, List

from modelscope.metrics.base import Metric
from modelscope.metrics.builder import METRICS, MetricKeys
from modelscope.utils.tensor_utils import torch_nested_detach, torch_nested_numpify

from adaseq.data.constant import NONE_REL_LABEL
from adaseq.metainfo import Metrics


def compute_f1(preds: List[str], labels: List[str]) -> Dict[str, float]:
    """Compute f1 from predictions and ground truth

    Args:
        preds (List[str]): prediction list
        labels (List[str]): ground truth list

    Returns:
        A dict containing: precision, recall, f1-score
    """
    # From https://github.com/princeton-nlp/PURE/blob/main/run_relation.py
    n_gold = n_pred = n_correct = 0
    for pred, label in zip(preds, labels):
        if pred != NONE_REL_LABEL:
            n_pred += 1
        if label != NONE_REL_LABEL:
            n_gold += 1
        if (pred != NONE_REL_LABEL) and (label != NONE_REL_LABEL) and (pred == label):
            n_correct += 1
    if n_correct == 0:
        return {'precision': 0.0, 'recall': 0.0, 'f1': 0.0}
    else:
        prec = n_correct * 1.0 / n_pred
        recall = n_correct * 1.0 / n_gold
        if prec + recall > 0:
            f1 = 2.0 * prec * recall / (prec + recall)
        else:
            f1 = 0.0

        return {'precision': prec, 'recall': recall, 'f1': f1}


@METRICS.register_module(module_name=Metrics.relation_extraction_metric)
class RelationExtractionMetric(Metric):
    """The metric computation class for relation extraction tasks."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.preds = []
        self.golds = []

    def add(self, outputs: Dict, inputs: Dict):
        """Collect batch outputs"""
        pred_results = outputs['predicts']
        ground_truths = inputs['label_id']
        self.preds.extend(torch_nested_numpify(torch_nested_detach(pred_results)).tolist())
        self.golds.extend(torch_nested_numpify(torch_nested_detach(ground_truths)).tolist())

    def evaluate(self):
        """Calculate metrics, returning precision, recall, f1-score in a dictionary

        Returns:
            scores (Dict):
                precision (float): micro averaged precision
                recall (float): micro averaged recall
                f1 (float): micro averaged f1-score
        """
        id2label = self.trainer.train_preprocessor.id_to_label

        pred_labels = [id2label[p] for p in self.preds]
        gold_labels = [id2label[p] for p in self.golds]

        overall_score = compute_f1(pred_labels, gold_labels)

        scores = {}
        scores[MetricKeys.PRECISION] = overall_score['precision']
        scores[MetricKeys.RECALL] = overall_score['recall']
        scores[MetricKeys.F1] = overall_score['f1']
        return scores
